#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation,
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""
SchoolTool Quiz CanDo views.
"""

from xml.sax.saxutils import quoteattr

from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.interface import implements
from zope.publisher.interfaces import NotFound
from zope.security import proxy
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.i18n import translate
import zc.table

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.cando.interfaces import IProjects
from schooltool.cando.interfaces import IProjectsGradebook
from schooltool.cando.interfaces import ISectionSkills
from schooltool.cando.interfaces import ISectionSkillSet
from schooltool.cando.interfaces import ISkillsGradebook
from schooltool.cando.browser.gradebook import CanDoGradeStudent
from schooltool.cando.browser.gradebook import CanDoGradeStudentTable
from schooltool.cando.browser.gradebook import CanDoGradeStudentTableFormatter
from schooltool.cando.browser.gradebook import MySkillsGradesView
from schooltool.cando.browser.gradebook import MySkillsGradesTable
from schooltool.cando.browser.gradebook import SkillSetColumn
from schooltool.cando.browser.gradebook import StudentPopupMenuView
from schooltool.cando.browser.gradebook import label_title_formatter
from schooltool.cando.skill import querySkillScoreSystem
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.course.interfaces import ISection
from schooltool.gradebook.browser.report_utils import buildHTMLParagraphs
from schooltool.gradebook.browser.gradebook import getScoreSystemDiscreteValues
from schooltool.gradebook.browser.gradebook import convertAverage
from schooltool.person.interfaces import IPerson
from schooltool.requirement.interfaces import ICommentScoreSystem
from schooltool.requirement.interfaces import IValuesScoreSystem
from schooltool.requirement.interfaces import IDiscreteValuesScoreSystem
from schooltool.requirement.interfaces import IEvaluations
from schooltool.requirement.scoresystem import UNSCORED
from schooltool.skin import flourish
from schooltool import table

from schooltool.quiz import interfaces
from schooltool.quiz import QuizMessage as _
from schooltool.quiz.browser.deployedquiz import (
    DeployedQuizAnswersViewTableViewlet)

from collections import defaultdict
from sys import maxint


class CanDoGradeStudentQuizInfo(CanDoGradeStudent):

    implements(interfaces.ISchoolToolQuizView)

    form_container_id = 'grade-with-quiz-info-container'
    recent_count_filter_id = 'recent-count-filter'
    default_recent_count = 1

    @Lazy
    def quiz_items(self):
        app = ISchoolToolApplication(None)
        return interfaces.IQuizItemContainer(app).values()

    @Lazy
    def recent_counts(self):
        result = []
        start = 1
        end = 11
        step = 1
        for value in range(start, end, step):
            result.append({
                    'value': value,
                    'submit_value': str(value),
                    'default': value == self.default_recent_count,
                    })
        return result

    @property
    def recent_count(self):
        recent_count_filter_id = self.recent_count_filter_id
        submitted_recent_count = self.request.get(recent_count_filter_id)
        if submitted_recent_count is not None:
            for option in self.recent_counts:
                if submitted_recent_count == option['submit_value']:
                    return option['value']
        return self.default_recent_count


class CanDoGradeStudentQuizInfoTableFormatter(CanDoGradeStudentTableFormatter):

    def renderCell(self, item, column):
        klass = self.cssClasses.get('td', '')
        if column.name == 'student-score':
            if klass:
                klass += ' '
            klass += 'student-score'
        if column.name == 'items':
            if klass:
                klass += ' '
            klass += 'items'
        if column.name == 'recent':
            if klass:
                klass += ' '
            klass += 'recent'
        if column.name == 'correct':
            if klass:
                klass += ' '
            klass += 'correct'
        if column.name == 'skill' and not item['skill'].required:
            if klass:
                klass += ' '
            klass += 'optional'
        klass = klass and ' class=%s' % quoteattr(klass) or ''
        return '<td id="%s"%s>%s</td>' % (
            item['skill_id'], klass, self.getCell(item, column),)


def skill_items_getter(view):
    def formatter(item, formatter):
        evaluations = IEvaluations(view.student)
        skill = item['skill']
        equivalent = skill.findAllEquivalent()
        result = []
        for item in view.quiz_items:
            if item.skill in equivalent:
                score = evaluations.get(item)
                if score is not None:
                    result.append((item, score))
        return result
    return formatter


def skill_items_formatter(view):
    def formatter(items, item, formatter):
        count = len(items)
        if count:
            student = formatter.context.student
            skill = item['skill']
            skillset = item['skillset']
            url = '%s/gradebook/%s/answers?skill=%s' % (
                absoluteURL(skillset, formatter.request),
                student.__name__,
                skill.__name__)
            return '<a href="%s" target="new">%s</a>' % (url, count)
        return ''
    return formatter


def skill_correct_formatter(items, item, formatter):
    correct = []
    for item, score in items:
        if score.value is not UNSCORED and score.value >= 100:
            correct.append(1)
    try:
        result = len(correct) * 100.0 / len(items)
    except (ZeroDivisionError,):
        return ''
    return '%.0f%%' % result


def skill_recent_formatter(view):
    def formatter(items, item, formatter):
        correct = []
        items = sorted(items,
                       key=lambda (item, score): score.time,
                       reverse=True)
        recent_count = view.recent_count
        items = items[:recent_count]
        for item, score in items:
            if score.value is not UNSCORED and score.value >= 100:
                correct.append(1)
        try:
            result = len(correct) * 100.0 / len(items)
        except (ZeroDivisionError,):
            return ''
        return '%.0f%%' % result
    return formatter


class CanDoGradeStudentQuizInfoTable(CanDoGradeStudentTable):

    table_formatter = CanDoGradeStudentQuizInfoTableFormatter
    visible_column_names = ['skill', 'items', 'correct',
                            'recent', 'student-score']

    @property
    def html_id(self):
        return self.view.form_container_id

    def columns(self):
        default = super(CanDoGradeStudentQuizInfoTable, self).columns()
        skillset, skill, score = default
        items = zc.table.column.GetterColumn(
            name='items',
            title=_('# of Items'),
            getter=skill_items_getter(self.view),
            cell_formatter=skill_items_formatter(self.view))
        correct = zc.table.column.GetterColumn(
            name='correct',
            title=_('% Correct Items'),
            getter=skill_items_getter(self.view),
            cell_formatter=skill_correct_formatter)
        recent = zc.table.column.GetterColumn(
            name='recent',
            title=_('% Correct Most Recent'),
            getter=skill_items_getter(self.view),
            cell_formatter=skill_recent_formatter(self.view))
        return [skillset, skill, items, correct, recent, score]


class RecentCountFilterViewlet(flourish.page.RefineLinksViewlet):

    pass


class RecentCountFilterMenuViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <select tal:attributes="id view/view/recent_count_filter_id;
                              name view/view/recent_count_filter_id;"
              class="select-widget report-filter">
        <option tal:repeat="option view/options"
                tal:content="option/title"
                tal:attributes="value option/value;
                                selected option/selected;"
                />
      </select>
    ''')

    def options(self):
        result = []
        for recent_count in self.view.recent_counts:
            selected = recent_count['value'] == self.view.recent_count
            result.append({
                    'value': recent_count['submit_value'],
                    'title': recent_count['value'],
                    'selected': selected,
                    })
        return result


class SkillAnswersView(flourish.page.Page):

    implements(interfaces.ISchoolToolQuizView,
               interfaces.IMathJaxView)

    @property
    def title(self):
        return self.context.student.title

    @property
    def subtitle(self):
        return label_title_formatter(self.skill, None, None)

    @Lazy
    def gradebook(self):
        return proxy.removeSecurityProxy(self.context.gradebook)

    @Lazy
    def skillset(self):
        return self.gradebook.context

    @Lazy
    def skill(self):
        skill_id = self.request.get('skill')
        skill = self.skillset.get(skill_id)
        if skill is None:
            raise NotFound(self.context, skill_id, self.request)
        return skill

    @Lazy
    def student(self):
        return self.context.student

    @Lazy
    def quiz_items(self):
        app = ISchoolToolApplication(None)
        return interfaces.IQuizItemContainer(app).values()


class SkillAnswersViewTableViewlet(DeployedQuizAnswersViewTableViewlet):

    template = ViewPageTemplateFile('templates/answers_skill.pt')

    def evaluations(self):
        person = self.view.student
        evaluations = IEvaluations(proxy.removeSecurityProxy(person))
        equivalent = self.view.skill.findAllEquivalent()
        result = []
        for item in self.view.quiz_items:
            if item.skill in equivalent:
                score = evaluations.get(item)
                if score is not None:
                    result.append(score)
        return sorted(result,
                      key=lambda score: score.time,
                      reverse=True)


# XXX: Leave this here until popup menus become viewlet managers
class StudentPopupQuizInfoMenuView(StudentPopupMenuView):

    def options(self, student):
        default = super(StudentPopupQuizInfoMenuView, self).options(student)
        gradeurl = '%s/%s/grade_with_quiz_info' % (
            absoluteURL(self.context, self.request),
            student.username)
        default.insert(2, {
                'label': self.translate(_('Score with Quiz Info')),
                'url': gradeurl,
                })
        return default


class MySkillsGradesViewQuizInfo(MySkillsGradesView):

    implements(interfaces.ISchoolToolQuizView)

    container_class = 'container widecontainer'
    form_container_id = 'grade-with-quiz-info-container'
    recent_count_filter_id = 'recent-count-filter'
    default_recent_count = 1

    @Lazy
    def student(self):
        person = IPerson(self.request.principal)
        return proxy.removeSecurityProxy(person)

    @Lazy
    def quiz_items(self):
        app = ISchoolToolApplication(None)
        return interfaces.IQuizItemContainer(app).values()

    @Lazy
    def recent_counts(self):
        result = []
        start = 1
        end = 11
        step = 1
        for value in range(start, end, step):
            result.append({
                    'value': value,
                    'submit_value': str(value),
                    'default': value == self.default_recent_count,
                    })
        return result

    @property
    def recent_count(self):
        recent_count_filter_id = self.recent_count_filter_id
        submitted_recent_count = self.request.get(recent_count_filter_id)
        if submitted_recent_count is not None:
            for option in self.recent_counts:
                if submitted_recent_count == option['submit_value']:
                    return option['value']
        return self.default_recent_count

    @Lazy
    def isSkillsGradebook(self):
        return ISectionSkillSet.providedBy(proxy.removeSecurityProxy(self.context).context)

    @property
    def worksheets(self):
        section = ISection(proxy.removeSecurityProxy(self.context).context)
        if self.isSkillsGradebook:
            return ISectionSkills(section)
        else:
            return IProjects(section)

    # XXX: almost duplicated from MyGradesView
    def update(self):
        self.person = IPerson(self.request.principal)
        self.processColumnPreferences()
        self.average = None
        self.table = []
        for worksheet in self.worksheets.values():
            if self.isSkillsGradebook:
                gradebook = ISkillsGradebook(worksheet)
                skillset_label = worksheet.label
            else:
                gradebook = IProjectsGradebook(worksheet)
                skillset_label = None
            for activity in gradebook.activities:
                activity = proxy.removeSecurityProxy(activity)
                score = gradebook.getScore(self.person, activity)
                if score:
                    if ICommentScoreSystem.providedBy(score.scoreSystem):
                        grade = {
                            'comment': True,
                            'paragraphs': buildHTMLParagraphs(score.value),
                            }
                    elif IValuesScoreSystem.providedBy(score.scoreSystem):
                        s_min, s_max = getScoreSystemDiscreteValues(score.scoreSystem)
                        value = score.value
                        if IDiscreteValuesScoreSystem.providedBy(score.scoreSystem):
                            value = score.scoreSystem.getNumericalValue(score.value)
                            if value is None:
                                value = 0
                        if int(value) != value:
                            value = '%.1f' % value
                        grade = {
                            'comment': False,
                            'value': '%s / %s' % (value, score.scoreSystem.getBestScore()),
                            }

                    else:
                        grade = {
                            'comment': False,
                            'value': score.value,
                            }

                else:
                    grade = {
                        'comment': False,
                        'value': '',
                        }

                title = activity.title
                if activity.description:
                    title += ' - %s' % activity.description

                row = {
                    'activity': title,
                    'grade': grade,
                    'object': activity,
                    'skill': activity,
                    'skillset': proxy.removeSecurityProxy(worksheet),
                    'skillset_label': skillset_label,
                    }
                self.table.append(row)

        """Handle change of current term."""
        if self.handleTermChange():
            return

        """Handle change of current section."""
        self.handleSectionChange()


def mygrades_skill_items_getter(view):
    def formatter(item, formatter):
        evaluations = IEvaluations(view.student)
        skill = item['object']
        equivalent = skill.findAllEquivalent()
        result = []
        for item in view.quiz_items:
            if item.skill in equivalent:
                score = evaluations.get(item)
                if score is not None:
                    result.append((item, score))
        return result
    return formatter


def mygrades_skill_items_formatter(view):
    def formatter(items, item, formatter):
        count = len(items)
        if count:
            skill = item['object']
            skillset = item['skillset']
            url = '%s/mygrades/answers?skill=%s' % (
                absoluteURL(skillset, formatter.request),
                skill.__name__)
            return '<a href="%s" target="new">%s</a>' % (url, count)
        return ''
    return formatter


def mygrades_skill_correct_formatter(items, item, formatter):
    correct = []
    for item, score in items:
        if score.value is not UNSCORED and score.value >= 100:
            correct.append(1)
    try:
        result = len(correct) * 100.0 / len(items)
    except (ZeroDivisionError,):
        return ''
    return '%.0f%%' % result


def mygrades_skill_recent_formatter(view):
    def formatter(items, item, formatter):
        correct = []
        items = sorted(items,
                       key=lambda (item, score): score.time,
                       reverse=True)
        recent_count = view.recent_count
        items = items[:recent_count]
        for item, score in items:
            if score.value is not UNSCORED and score.value >= 100:
                correct.append(1)
        try:
            result = len(correct) * 100.0 / len(items)
        except (ZeroDivisionError,):
            return ''
        return '%.0f%%' % result
    return formatter


class MyGradesTableQuizInfoTableFormatter(table.ajax.AJAXFormSortFormatter):

    def renderCell(self, item, column):
        klass = self.cssClasses.get('td', '')
        if column.name == 'score':
            if klass:
                klass += ' '
            klass += 'score'
        if column.name == 'items':
            if klass:
                klass += ' '
            klass += 'items'
        if column.name == 'recent':
            if klass:
                klass += ' '
            klass += 'recent'
        if column.name == 'correct':
            if klass:
                klass += ' '
            klass += 'correct'
        if column.name == 'skill' and not item['object'].required:
            if klass:
                klass += ' '
            klass += 'optional'
        klass = klass and ' class=%s' % quoteattr(klass) or ''
        skill_id = self.getSkillId(item['object'])
        return '<td id="%s"%s>%s</td>' % (
            skill_id, klass, self.getCell(item, column),)

    def getSkillId(self, skill):
        skillset = skill.__parent__
        return '%s.%s' % (skillset.__name__, skill.__name__)


class MyGradesTableQuizInfo(MySkillsGradesTable):

    batch_size = 0
    group_by_column = 'skillset'
    table_formatter = MyGradesTableQuizInfoTableFormatter
    visible_column_names = ['skill', 'items', 'correct', 'recent', 'score']

    def columns(self):
        default = super(MyGradesTableQuizInfo, self).columns()
        skill_sorting, skill, score = default
        skillset = SkillSetColumn(
            name='skillset',
            title=_('Skill Set'))
        items = zc.table.column.GetterColumn(
            name='items',
            title=_('# of Items'),
            getter=mygrades_skill_items_getter(self.view),
            cell_formatter=mygrades_skill_items_formatter(self.view))
        correct = zc.table.column.GetterColumn(
            name='correct',
            title=_('% Correct Items'),
            getter=mygrades_skill_items_getter(self.view),
            cell_formatter=mygrades_skill_correct_formatter)
        recent = zc.table.column.GetterColumn(
            name='recent',
            title=_('% Correct Most Recent'),
            getter=mygrades_skill_items_getter(self.view),
            cell_formatter=mygrades_skill_recent_formatter(self.view))
        return [skillset, skill, items, correct, recent, score]

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'grade-student'})

    def sortOn(self):
        return (('skillset', False),)


class MyGradesSkillAnswersView(flourish.page.Page):

    implements(interfaces.ISchoolToolQuizView,
               interfaces.IMathJaxView)

    @property
    def title(self):
        return IPerson(self.request.principal).title

    @property
    def subtitle(self):
        return label_title_formatter(self.skill, None, None)

    @Lazy
    def skillset(self):
        return self.skill.__parent__

    @Lazy
    def skill(self):
        skill_id = self.request.get('skill')
        skill = proxy.removeSecurityProxy(self.context).context.get(skill_id)
        if skill is None:
            raise NotFound(self.context, skill_id, self.request)
        return skill

    @Lazy
    def student(self):
        return IPerson(self.request.principal)

    @Lazy
    def quiz_items(self):
        app = ISchoolToolApplication(None)
        return interfaces.IQuizItemContainer(app).values()


class UpdateSkillsLink(flourish.page.ModalFormLinkViewlet):
    
    @property
    def enabled(self):
        gradebook = proxy.removeSecurityProxy(self.context)
        app = ISchoolToolApplication(None)
        preferences = interfaces.IApplicationPreferences(app)
        section = gradebook.section
        deployed_quizzes = interfaces.IDeployedQuizContainer(section)
        return ISkillsGradebook.providedBy(gradebook) and \
               preferences.skill_rules and \
               deployed_quizzes

    @property
    def dialog_title(self):
        title = _(u'Update Skills')
        return translate(title, context=self.request)



class UpdateSkillsView(flourish.form.Dialog):

    def update(self):
        super(UpdateSkillsView, self).update() 
        if 'DONE' in self.request:
            url = absoluteURL(self.context, self.request)
            self.request.response.redirect(url)
            return
        app = ISchoolToolApplication(None)
        rules = {}
        Rules = interfaces.IApplicationPreferences(app)
        for rule in Rules.skill_rules:
            rules[rule.level] = (rule.total_taken, rule.out_of , rule.number_correct)
            newrules = ()
            for rulepart in rules[rule.level]:
                if rulepart:
                   newrules = newrules+(rulepart,)
                else:
                   newrules = newrules+(maxint,)
            rules[rule.level] = newrules
        maxlevel = max(querySkillScoreSystem().scoresDict.keys())
        worksheets = self.context.__parent__.__parent__
        section = proxy.removeSecurityProxy(worksheets.__parent__)
        memberlist = [member for member in section.members]
        quizitems = interfaces.IQuizItemContainer(app)
        for skillset in worksheets.values():
            gradebook = ISkillsGradebook(skillset)
            worksheetskills = {}
            worksheetequivalentskills = {}
            equivalent_to_worksheet = {}
            student_to_skills = {}
            evaluated_skills = []
            for skill in skillset.values():
                worksheetskills[skill] = None
                for equivskill in skill.findAllEquivalent():
                    worksheetequivalentskills[equivskill] = None
                equivalent_to_worksheet[tuple(skill.findAllEquivalent())] = skill
            for quizitem in quizitems.values():
                if quizitem.skill not in evaluated_skills:
                    evaluated_skills.append(quizitem.skill)
            for student in memberlist:
                for skill in worksheetskills:
                    worksheetskills[skill] = None
                evaluations = IEvaluations(proxy.removeSecurityProxy(student))
                relevant_evaluations = defaultdict(list)
                for quizitem in quizitems.values():
                    if quizitem.skill in worksheetequivalentskills.keys():
                        question = evaluations.get(quizitem)
                        if question is not None:
                            relevant_evaluations[quizitem.skill].append(question)
                for skill in relevant_evaluations:
                    relevant_evaluations[skill] = sorted(relevant_evaluations[skill],
                                                      key=lambda question: question.time,
                                                      reverse=True)
                skill_to_level = {}
                for skill in evaluated_skills:
                    skill_to_level[skill] = self.calculate_skill_level(rules, relevant_evaluations[skill])
                student_to_skills[student] = skill_to_level
                for skill in evaluated_skills:
                    for key in equivalent_to_worksheet.keys():
                        if skill in key:
                            worksheetskills[equivalent_to_worksheet[key]] = skill_to_level[skill]
                for skill in worksheetskills:
                    if worksheetskills[skill] == None:
                        worksheetskills[skill] = UNSCORED
                    scoresystem = proxy.removeSecurityProxy(skill.scoresystem)
                    current_score = gradebook.getScore(student, skill)
                    if (current_score and current_score.value == maxlevel and \
                        worksheetskills[skill] == (int(maxlevel)-1)) or (current_score and \
                        current_score.evaluator is not None):
                        pass
                    elif worksheetskills[skill] == UNSCORED:
                        gradebook.evaluate(student,
                                           skill,
                                           UNSCORED,
                                           None)
                    else:
                        gradebook.evaluate(student,
                                           skill,
                                           str(scoresystem.scoresDict[str(worksheetskills[skill])]), 
                                           None) 
        self.status = _("Skills successfully updated!") 

    def calculate_skill_level(self, rules, itemlist):
        level = 0
        for rulelevel,ruleset in zip(rules.keys(),rules.values()):
            total_taken = ruleset[0]
            out_of = ruleset[1]
            number_correct = ruleset[2]
            most_recent = itemlist[:out_of]
            most_recent_passed = [item.value for item in most_recent].count(100)
            if (len(itemlist) >= total_taken or
                most_recent_passed >= number_correct):
                level = rulelevel
        return level

    def initDialog(self):
        super(UpdateSkillsView, self).initDialog()
        self.ajax_settings['dialog']['dialogClass'] = 'explicit-close-dialog'
        self.ajax_settings['dialog']['closeOnEscape'] = False
        self.ajax_settings['dialog']['modal'] = True


class MyGradesRecentCountFilterMenuViewlet(RecentCountFilterMenuViewlet):

    template = InlineViewPageTemplate('''
      <select tal:attributes="id view/view/recent_count_filter_id;
                              name view/view/recent_count_filter_id;"
              class="select-widget report-filter"
              onchange="ST.redirect($(this).context.value);">
        <option tal:repeat="option view/options"
                tal:content="option/title"
                tal:attributes="value string:?${view/view/recent_count_filter_id}=${option/value};
                                selected option/selected;"
                />
      </select>
    ''')
