#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation,
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""
SchoolTool Quiz deployed quiz views.
"""

from decimal import Decimal
import random
from datetime import datetime, timedelta
from zope import schema
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.component import getUtility
from zope.container.interfaces import INameChooser
from zope.dublincore.interfaces import IZopeDublinCore
from zope.i18n import translate
from zope.i18n.interfaces.locales import ICollator
from zope.interface import implements, directlyProvides
from zope.publisher.interfaces import NotFound
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm
from zope.schema.interfaces import IVocabularyFactory
from zope.security import checkPermission
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.absoluteurl import absoluteURL

from z3c.form import datamanager, widget
from z3c.form import field, form, button
from z3c.form.interfaces import DISPLAY_MODE, HIDDEN_MODE
from zc.table.column import GetterColumn
from zc.table.interfaces import ISortableColumn

from schooltool import table
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.utils import vocabulary
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.course.interfaces import IInstructor, ISection
from schooltool.course.interfaces import ILearner
from schooltool.gradebook.interfaces import IActivities
from schooltool.gradebook.interfaces import IGradebook
from schooltool.gradebook.interfaces import ICategoryContainer
from schooltool.gradebook.activity import ensureAtLeastOneWorksheet
from schooltool.gradebook.activity import Activity
from schooltool.person.interfaces import IPerson
from schooltool.lyceum.journal.interfaces import ISectionJournal
from schooltool.requirement.interfaces import IEvaluations
from schooltool.requirement.scoresystem import UNSCORED
from schooltool.requirement.scoresystem import PercentScoreSystem
from schooltool.requirement.evaluation import Evaluation
from schooltool.skin import flourish
from schooltool.term.interfaces import ITerm, IDateManager
from schooltool.requirement.scoresystem import RangedValuesScoreSystem

from schooltool.quiz import QuizMessage as _
from schooltool.quiz import interfaces
from schooltool.quiz.deployedquiz import DeployedQuiz
from schooltool.quiz.assignments import Assignment
from schooltool.quiz.browser.quiz import taken
from schooltool.quiz.browser.widget import LabeledCheckBoxFieldWidget
from schooltool.quiz.browser.widget import PreLabeledCheckBoxFieldWidget
from schooltool.quiz.browser.widget import RestLabeledCheckBoxFieldWidget
from schooltool.quiz.browser.widget import RestLabeledRadioFieldWidget
from schooltool.quiz.utils import render_datetime, format_percentage
from schooltool.quiz.utils import render_rst, render_pre
from schooltool.quiz.browser.quiz import SolutionDoneLinkViewlet
from schooltool.quiz.utils import render_date


def getStartTime(adapter):
    today = getUtility(IDateManager).today
    return datetime(today.year, today.month, today.day)


def getEndTime(adapter):
    time = getStartTime(adapter)
    return time + timedelta(days=1)


DeployedQuizDefaultTitle = widget.ComputedWidgetAttribute(
    lambda adapter: adapter.context.title,
    field=interfaces.IDeployedQuiz['title']
    )


DeployedQuizDefaultStart = widget.ComputedWidgetAttribute(
    getStartTime,
    field=interfaces.IDeployedQuiz['start']
    )


DeployedQuizDefaultEnd = widget.ComputedWidgetAttribute(
    getEndTime,
    field=interfaces.IDeployedQuiz['end']
    )


class DeployedQuizForm(flourish.form.Form):

    implements(interfaces.ISchoolToolQuizView)

    id = 'quiz-deploy-form'
    legend = _('Quiz Deployment Information')
    content_template = ViewPageTemplateFile('templates/form.pt')


class QuizDeployViewBase(object):

    @property
    def sections(self):
        result = []
        current_term = getUtility(IDateManager).current_term
        person = IPerson(self.request.principal, None)
        if person is not None:
            for section in list(IInstructor(person).sections()):
                term = ITerm(section)
                if term == current_term:
                    result.append(section)
        return result

    @property
    def section_vocabulary(self):
        collator = ICollator(self.request.locale)
        terms = []
        for section in sorted(self.sections,
                              key=lambda section: section.title,
                              cmp=collator.cmp):
            term = SimpleTerm(section,
                              token=section.__name__,
                              title=section.title)
            terms.append(term)
        return SimpleVocabulary(terms)

    @property
    def section_field(self):
        return schema.Choice(
            __name__='section',
            title=_('Section'),
            vocabulary=self.section_vocabulary)

    @property
    def deployment_types_vocabulary(self):
        collator = ICollator(self.request.locale)
        items = [
            ('gradebook', _('Gradebook')),
            ('journal', _('Journal')),
            ]
        terms = []
        for key, title in sorted(items,
                                 key=lambda (key, title): title,
                                 cmp=collator.cmp):
            term = SimpleTerm(title, token=key, title=title)
            terms.append(term)
        return SimpleVocabulary(terms)

    @property
    def gradebook_worksheets_vocabulary(self):
        collator = ICollator(self.request.locale)
        terms = []
        for section in self.sections:
            worksheets = IActivities(section)
            ensureAtLeastOneWorksheet(worksheets)
            for worksheet in worksheets.values():
                if not worksheet.deployed:
                    terms.append(
                        SimpleTerm(worksheet,
                                   token='%s.%s' % (section.__name__,
                                                    worksheet.__name__),
                                   title=worksheet.title))
        terms = sorted(terms,
                       key=lambda term: term.title,
                       cmp=collator.cmp)
        return SimpleVocabulary(terms)

    @property
    def deployment_type_field(self):
        return schema.Choice(
            __name__='deployment_type',
            title=_('Submit Grades To'),
            vocabulary=self.deployment_types_vocabulary)

    @property
    def worksheets_field(self):
        return schema.Choice(
            __name__='worksheet',
            title=_('Worksheet'),
            vocabulary=self.gradebook_worksheets_vocabulary,
            required=True)

    @property
    def periods_field(self):
        return schema.Date(
            __name__='period',
            title=_('Journal Period'),
            required=False)

    def getDefaultCategory(self):
        categories = ICategoryContainer(ISchoolToolApplication(None))
        return categories.default_key

    def addActivity(self, deployed, activity_title=None):
        scoresystem = RangedValuesScoreSystem(u'generated', min=0, max=100)
        category = self.getDefaultCategory()
        worksheets = IActivities(self._section)
        worksheet = removeSecurityProxy(worksheets[deployed.worksheet_id])
        if activity_title is None:
            activity_title = deployed.title
        activity = Activity(activity_title, category, scoresystem)
        chooser = INameChooser(worksheet)
        name = chooser.chooseName('', activity)
        worksheet[name] = activity
        deployed.activity_id = name


class QuizDeployView(DeployedQuizForm,
                     flourish.form.AddForm,
                     QuizDeployViewBase):

    @property
    def fields(self):
        fields = field.Fields(self.section_field)
        fields += field.Fields(self.deployment_type_field)
        fields += field.Fields(self.worksheets_field)
        fields += field.Fields(self.periods_field)
        fields += field.Fields(interfaces.IDeployedQuiz).select('title',
                                                                'random')
        fields += field.Fields(interfaces.IDeployedQuiz['open_by_teacher'],
                               mode=HIDDEN_MODE)
        fields['random'].widgetFactory = LabeledCheckBoxFieldWidget
        return fields

    def updateActions(self):
        super(QuizDeployView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def create(self, data):
        self._section = data.pop('section')
        worksheet = data.pop('worksheet')
        period = data.pop('period')
        deployment_type = data.pop('deployment_type')
        term = self.deployment_types_vocabulary.getTerm(deployment_type)
        self._deployment_type = term.token
        if self._deployment_type == 'gradebook':
            data['worksheet_id'] = worksheet.__name__
        else:
            journal = ISectionJournal(self._section)
            result = [meeting for meeting in journal.meetings
                      if meeting.dtstart.year == period.year and
                      meeting.dtstart.month == period.month and
                      meeting.dtstart.day == period.day]
            if result:
                meeting = result[0]
                data['period_id'] = meeting.unique_id
            else:
                # raise validation error
                pass
        deployed = DeployedQuiz()
        form.applyChanges(self, deployed, data)
        if self._deployment_type == 'gradebook':
            deployed.worksheet_id = data['worksheet_id']
        else:
            try:
                deployed.period_id = data['period_id']
            except:
                raise schema.ValidationError('There is no meeting in '
                                             'the journal for this date')
        return deployed

    def add(self, deployed):
        container = interfaces.IDeployedQuizContainer(self._section)
        chooser = INameChooser(container)
        name = chooser.chooseName('', deployed)
        container[name] = deployed
        deployed.quizzes.add(removeSecurityProxy(self.context))
        for item in self.context:
            deployed.items.add(removeSecurityProxy(item))
            deployed.order.append(item.__name__)
        self._deployed = deployed
        if self._deployment_type == 'gradebook':
            self.addActivity(deployed)
        return deployed

    def nextURL(self):
        if self._finishedAdd:
            url = absoluteURL(self._deployed, self.request)
        else:
            url = self.request.get('camefrom',
                                   absoluteURL(self.context, self.request))
        return url

    def updateWidgets(self):
        super(QuizDeployView, self).updateWidgets()
        self.widgets['section'].prompt = True
        self.widgets['section'].promptMessage = _('Select a section')
        self.widgets['deployment_type'].prompt = True
        self.widgets['deployment_type'].promptMessage = _('Select a target')


class DeployedQuizView(flourish.page.Page):

    implements(interfaces.ISchoolToolQuizView)

    @property
    def title(self):
        return self.context.title

    @property
    def subtitle(self):
        return self.section.title

    @Lazy
    def section(self):
        return ISection(self.context)

    @property
    def can_modify(self):
        return checkPermission('schooltool.edit', self.context)

    @property
    def now(self):
        today = getUtility(IDateManager).today
        now = datetime.now()
        return datetime(today.year, today.month, today.day,
                        now.hour, now.minute)

    @property
    def is_open(self):
        return self.context.open_by_teacher


class DeployedQuizEditView(DeployedQuizForm, form.EditForm):

    fields = field.Fields(interfaces.IDeployedQuiz).select('title',
                                                           'random')
    fields['random'].widgetFactory = LabeledCheckBoxFieldWidget

    def update(self):
        return form.EditForm.update(self)

    def updateActions(self):
        super(DeployedQuizEditView, self).updateActions()
        self.actions['submit'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def nextURL(self):
        camefrom = self.request.get('camefrom')
        url = absoluteURL(self.context, self.request)
        if camefrom is not None:
            return '%s?camefrom=%s' % (url, camefrom)
        return url

    @button.buttonAndHandler(_('Submit'), name='submit')
    def handle_submit(self, action):
        super(DeployedQuizEditView, self).handleApply.func(self, action)
        if (self.status == self.successMessage or
            self.status == self.noChangesMessage):
            self.request.response.redirect(self.nextURL())

    @button.buttonAndHandler(_('Cancel'), name='cancel')
    def handle_cancel(self, action):
        self.request.response.redirect(self.nextURL())


class DeployedQuizGradingTargetEditView(DeployedQuizForm,
                                        form.EditForm,
                                        QuizDeployViewBase):

    legend = _('Grading Target Information')

    def __init__(self, *args, **kw):
        super(DeployedQuizGradingTargetEditView, self).__init__(*args, **kw)
        section = ISection(self.context)
        self.values = {
            'worksheet': None,
            'activity': None,
            'period': None,
            }
        if self.context.period_id is None:
            self._deployment_type = 'gradebook'
            worksheets = IActivities(section)
            proxied_worksheet = worksheets.get(self.context.worksheet_id)
            worksheet = removeSecurityProxy(proxied_worksheet)
            if (worksheet is not None and not worksheet.deployed and
                not worksheet.hidden):
                self.values['worksheet'] = worksheet
                activity = worksheet.get(self.context.activity_id)
                if activity is not None:
                    self.values['activity'] = activity.title
        else:
            self._deployment_type = 'journal'
            journal = ISectionJournal(section)
            result = [meeting for meeting in journal.meetings
                      if meeting.unique_id == self.context.period_id]
            if result:
                self.values['period'] = result[0].dtstart.date()
        self.values['deployment_type'] = self._deployment_type
        self.values['section'] = section

    def getContent(self):
        return self.values

    @property
    def worksheets_field(self):
        return schema.Choice(
            __name__='worksheet',
            title=_('Worksheet'),
            vocabulary=self.gradebook_worksheets_vocabulary,
            required=False)

    @property
    def activity_field(self):
        return schema.TextLine(
            __name__='activity',
            title=_('Activity'),
            required=False)

    @property
    def fields(self):
        fields = field.Fields(self.section_field)
        fields += field.Fields(self.deployment_type_field)
        fields += field.Fields(self.worksheets_field)
        fields += field.Fields(self.activity_field)
        fields += field.Fields(self.periods_field)
        return fields

    def update(self):
        return form.EditForm.update(self)

    def updateWidgets(self):
        super(DeployedQuizGradingTargetEditView, self).updateWidgets()
        self.widgets['worksheet'].prompt = True
        self.widgets['worksheet'].promptMessage = _('Select a worksheet')

    def updateActions(self):
        super(DeployedQuizGradingTargetEditView, self).updateActions()
        self.actions['submit'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def nextURL(self):
        camefrom = self.request.get('camefrom')
        url = absoluteURL(self.context, self.request)
        if camefrom is not None:
            return '%s?camefrom=%s' % (url, camefrom)
        return url

    def detectChanges(self, data):
        return True

    @button.buttonAndHandler(_('Submit'), name='submit')
    def handle_submit(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        changes = self.detectChanges(data)
        if not changes:
            self.request.response.redirect(self.nextURL())
            return
        self._section = ISection(self.context)
        worksheet = data.pop('worksheet')
        period = data.pop('period')
        deployment_type = data.pop('deployment_type')
        term = self.deployment_types_vocabulary.getTerm(deployment_type)
        deployment_type = term.token
        if deployment_type == 'gradebook':
            worksheet_id = getattr(worksheet, '__name__')
            activity_title = data.pop('activity')
            if worksheet_id is not None:
                self.context.worksheet_id = worksheet_id
                self.addActivity(self.context, activity_title)
            data['period_id'] = None
        else:
            journal = ISectionJournal(self._section)
            result = [meeting for meeting in journal.meetings
                      if meeting.dtstart.year == period.year and
                      meeting.dtstart.month == period.month and
                      meeting.dtstart.day == period.day]
            if result:
                meeting = result[0]
                data['period_id'] = meeting.unique_id
                data['worksheet_id'] = None
                data['activity_id'] = None
        self.applyChanges(data)
        self.request.response.redirect(self.nextURL())

    @button.buttonAndHandler(_('Cancel'), name='cancel')
    def handle_cancel(self, action):
        self.request.response.redirect(self.nextURL())


class DeployedQuizDeleteView(flourish.form.DialogForm):

    template = ViewPageTemplateFile('templates/confirm_delete_deployed.pt')

    dialog_submit_actions = ('delete',)
    dialog_close_actions = ('cancel',)
    label = None

    @button.buttonAndHandler(_('Delete'), name='delete')
    def handle_delete(self, action):
        url = '%s/delete.html?delete.%s&CONFIRM' % (
            absoluteURL(self.context.__parent__, self.request),
            self.context.__name__)
        camefrom = self.request.get('camefrom')
        if camefrom is not None:
            url += '&camefrom=%s' % camefrom
        self.request.response.redirect(url)
        self.ajax_settings['dialog'] = 'close'

    @button.buttonAndHandler(_('Cancel'), name='cancel')
    def handle_cancel(self, action):
        pass

    def updateActions(self):
        super(DeployedQuizDeleteView, self).updateActions()
        self.actions['delete'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class DeployedQuizContainerDeleteView(flourish.containers.ContainerDeleteView):

    def nextURL(self):
        if 'CONFIRM' in self.request:
            camefrom = self.request.get('camefrom')
            if camefrom is not None:
                return camefrom
            person = IPerson(self.request.principal)
            url = '%s/quizzes.html' % absoluteURL(person, self.request)
            return url
        return super(DeployedQuizContainerDeleteView, self).nextURL()


class DeployedQuizzesViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <tal:block i18n:domain="schooltool.quiz"
                 tal:define="quizzes view/view/deployed_quizzes">
        <h3 i18n:translate="">Deployed Quizzes</h3>
        <div tal:condition="quizzes"
             tal:define="ajax nocall:view/view/providers/ajax"
             tal:content="structure ajax/view/context/deployed_quizzes_table"
             />
        <p i18n:translate="" tal:condition="not:quizzes">
          There are no quizzes deployed in any of your sections.
        </p>
      </tal:block>
    ''')

    @property
    def enabled(self):
        person = self.context
        instructor_sections = list(IInstructor(person).sections())
        return instructor_sections

    def render(self, *args, **kw):
        if self.enabled:
            return self.template(*args, **kw)


class DeploymentsViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <tal:block i18n:domain="schooltool.quiz"
                 tal:define="quizzes view/view/deployed_quizzes">
        <h3 i18n:translate="">Deployments of this quiz</h3>
        <div tal:condition="quizzes"
             tal:define="ajax nocall:view/view/providers/ajax"
             tal:content="structure ajax/view/context/deployed_quizzes_table"
             />
        <p i18n:translate="" tal:condition="not:quizzes">
          This quiz has not been deployed yet.
        </p>
      </tal:block>
    ''')


class DateTimeColumn(GetterColumn):

    implements(ISortableColumn)

    def getter(self, item, formatter):
        return getattr(item, self.name)

    def cell_formatter(self, value, item, formatter):
        return render_datetime(value)


def deployed_title_formatter(title, item, formatter):
    url = absoluteURL(item, formatter.request)
    camefrom = formatter.request.URL
    return '<a href="%s?camefrom=%s">%s</a>' % (url, camefrom, title)


def deployed_taken_getter(item, formatter):
    section = ISection(item)
    for student in section.members:
        if taken(student, item):
            return True
    return False


def deployed_taken_formatter(value, item, formatter):
    if value:
        return '<span class="ui-icon ui-icon-check">icon</span>'
    return ''


def deployed_closed_formatter(value, item, formatter):
    if not value:
        return '<span class="ui-icon ui-icon-check">icon</span>'
    return ''


class DeployedQuizzesTable(table.ajax.Table):

    def items(self):
        return self.view.deployed_quizzes

    def columns(self):
        title = table.column.LocaleAwareGetterColumn(
            name='title',
            title=_('Title'),
            getter=lambda deployed, formatter: deployed.title,
            cell_formatter=deployed_title_formatter)
        section = table.column.LocaleAwareGetterColumn(
            name='section',
            title=_('Section'),
            getter=lambda deployed, formatter: ISection(deployed).title)
        taken = GetterColumn(
            name='taken',
            title=_('Taken'),
            getter=deployed_taken_getter,
            cell_formatter=deployed_taken_formatter,
            subsort=True)
        directlyProvides(taken, ISortableColumn)
        created = GetterColumn(
            name='created',
            title=_('Created'),
            getter=lambda i, f: IZopeDublinCore(i).created,
            cell_formatter=lambda v, i, f: render_date(v),
            subsort=True)
        directlyProvides(created, ISortableColumn)
        closed = GetterColumn(
            name='closed',
            title=_('Closed'),
            getter=lambda deployed, formatter: deployed.open_by_teacher,
            cell_formatter=deployed_closed_formatter,
            subsort=True)
        directlyProvides(closed, ISortableColumn)
        return  [taken, title, section, created, closed]

    def sortOn(self):
        return (('created', True),)

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data deployed-quizzes-table'})


class TakeableQuizzesViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <tal:block i18n:domain="schooltool.quiz"
                 tal:define="quizzes view/view/takeable_quizzes">
        <h3 i18n:translate="">Quizzes to take</h3>
        <div tal:condition="quizzes"
             tal:define="ajax nocall:view/view/providers/ajax"
             tal:content="structure ajax/view/context/takeable_quizzes_table"
             />
        <p i18n:translate="" tal:condition="not:quizzes">
          There are no quizzes you can take.
        </p>
      </tal:block>
    ''')

    @property
    def enabled(self):
        person = self.context
        learner_sections = list(ILearner(person).sections())
        return learner_sections

    def render(self, *args, **kw):
        if self.enabled:
            return self.template(*args, **kw)


def takeable_deployed_title_formatter(title, item, formatter):
    if item.open_by_teacher:
        url = '%s/take.html' % absoluteURL(item, formatter.request)
        return '<a href="%s">%s</a>' % (url, title)
    return title


class TakeableQuizzesTable(table.ajax.Table):

    def items(self):
        return self.view.takeable_quizzes

    def columns(self):
        quiz = table.column.LocaleAwareGetterColumn(
            name='quiz',
            title=_('Quiz'),
            getter=lambda deployed, formatter: deployed.title,
            cell_formatter=takeable_deployed_title_formatter,
            subsort=True)
        section = table.column.LocaleAwareGetterColumn(
            name='section',
            title=_('Section'),
            getter=lambda deployed, formatter: ISection(deployed).title,
            subsort=True)
        created = GetterColumn(
            name='created',
            title=_('Created'),
            getter=lambda i, f: IZopeDublinCore(i).created,
            cell_formatter=lambda v, i, f: render_date(v),
            subsort=True)
        directlyProvides(created, ISortableColumn)
        closed = GetterColumn(
            name='closed',
            title=_('Closed'),
            getter=lambda deployed, formatter: deployed.open_by_teacher,
            cell_formatter=deployed_closed_formatter)
        return [quiz, section, created, closed]

    def sortOn(self):
        return (('created', True), ('section', False), ('quiz', False))

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data takeable-quizzes-table'})


class TakenQuizzesViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <tal:block i18n:domain="schooltool.quiz"
                 tal:define="quizzes view/view/taken_quizzes">
        <h3 i18n:translate="">Quizzes taken</h3>
        <div tal:condition="quizzes"
             tal:define="ajax nocall:view/view/providers/ajax"
             tal:content="structure ajax/view/context/taken_quizzes_table"
             />
        <p i18n:translate="" tal:condition="not:quizzes">
          You haven't taken any quizzes yet.
        </p>
      </tal:block>
    ''')

    @property
    def enabled(self):
        person = self.context
        learner_sections = list(ILearner(person).sections())
        return learner_sections

    def render(self, *args, **kw):
        if self.enabled:
            return self.template(*args, **kw)


def taken_deployed_title_formatter(title, item, formatter):
    if not item.open_by_teacher:
        student = formatter.context
        student_url = '%s/quizzes.html' % absoluteURL(student,
                                                      formatter.request)
        quiz_url = absoluteURL(item, formatter.request)
        username = formatter.context.username
        url = '%s/answers.html?student=%s&show_student=true&camefrom=%s' % (
            quiz_url, username, student_url)
        return '<a href="%s">%s</a>' % (url, title)
    return title


def student_grade(deployed, person):
    evaluations = IEvaluations(removeSecurityProxy(person))
    grades = []
    for item in deployed.items:
        item = removeSecurityProxy(item)
        evaluation = evaluations.get(item)
        if evaluation is not None:
            if evaluation.value is UNSCORED:
                # show grade only if all items are graded
                return _('N/A')
            else:
                grades.append(evaluation.value)
    grade = sum(grades) / float(len(grades))
    return format_percentage(grade)


def taken_grade_getter(deployed, formatter):
    return student_grade(deployed, formatter.context)


class TakenQuizzesTable(table.ajax.Table):

    def items(self):
        return self.view.taken_quizzes

    def columns(self):
        quiz = table.column.LocaleAwareGetterColumn(
            name='quiz',
            title=_('Quiz'),
            getter=lambda deployed, formatter: deployed.title,
            cell_formatter=taken_deployed_title_formatter)
        section = table.column.LocaleAwareGetterColumn(
            name='section',
            title=_('Section'),
            getter=lambda deployed, formatter: ISection(deployed).title)
        created = GetterColumn(
            name='created',
            title=_('Created'),
            getter=lambda i, f: IZopeDublinCore(i).created,
            cell_formatter=lambda v, i, f: render_date(v),
            subsort=True)
        directlyProvides(created, ISortableColumn)
        closed = GetterColumn(
            name='closed',
            title=_('Closed'),
            getter=lambda deployed, formatter: deployed.open_by_teacher,
            cell_formatter=deployed_closed_formatter)
        grade = GetterColumn(
            name='grade',
            title=_('Grade'),
            getter=taken_grade_getter)
        return [quiz, section, created, closed, grade]

    def sortOn(self):
        return (('created', True), ('section', False), ('quiz', False))

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data taken-quizzes-table'})


class DeployedQuizDetailsViewlet(flourish.form.FormViewlet):

    template = ViewPageTemplateFile('templates/deployed_details.pt')
    fields = field.Fields(interfaces.IDeployedQuiz).select('open_by_teacher',
                                                           'random')
    mode = DISPLAY_MODE

    def edit_url(self):
        context_url = absoluteURL(self.context, self.request)
        camefrom = self.request.get('camefrom')
        if camefrom is None:
            quiz = list(self.context.quizzes)[0]
            camefrom = absoluteURL(quiz, self.request)
        return '%s/edit.html?camefrom=%s' % (context_url, camefrom)


class DeployedQuizGradingDetailsViewlet(flourish.form.FormViewlet,
                                        QuizDeployViewBase):

    template = ViewPageTemplateFile('templates/deployed_grading_details.pt')
    mode = DISPLAY_MODE

    def __init__(self, *args, **kw):
        super(DeployedQuizGradingDetailsViewlet, self).__init__(*args, **kw)
        section = ISection(self.context)
        self.values = {}
        self.values['activity'] = _('Not found')
        self.values['worksheet'] = _('Not found')
        self.values['period'] = _('Not found')
        if self.context.period_id is None:
            self._deployment_type = 'gradebook'
            worksheets = IActivities(section)
            unproxied_worksheet = worksheets.get(self.context.worksheet_id)
            worksheet = removeSecurityProxy(unproxied_worksheet)
            if (worksheet is not None and not worksheet.deployed and
                not worksheet.hidden):
                self.values['worksheet'] = worksheet.title
                activity = worksheet.get(self.context.activity_id)
                if activity is not None:
                    self.values['activity'] = activity.title
        else:
            self._deployment_type = 'journal'
            journal = ISectionJournal(section)
            result = [meeting for meeting in journal.meetings
                      if meeting.unique_id == self.context.period_id]
            if result:
                self.values['period'] = result[0].dtstart.date()
        self.values['deployment_type'] = self._deployment_type

    def getContent(self):
        return self.values

    @property
    def fields(self):
        fields = field.Fields(self.deployment_type_field)
        if self._deployment_type == 'gradebook':
            schema_field = schema.TextLine(
                __name__='worksheet',
                title=_('Worksheet'))
            datamanager.DictionaryField(self.values, schema_field)
            fields += field.Fields(schema_field)
            schema_field = schema.TextLine(
                __name__='activity',
                title=_('Activity'))
            datamanager.DictionaryField(self.values, schema_field)
            fields += field.Fields(schema_field)
        else:
            schema_field = schema.Date(
                __name__='period',
                title=_('Period'))
            datamanager.DictionaryField(self.values, schema_field)
            fields += field.Fields(schema_field)
        return fields

    def edit_url(self):
        context_url = absoluteURL(self.context, self.request)
        camefrom = self.request.get('camefrom')
        if camefrom is None:
            quiz = list(self.context.quizzes)[0]
            camefrom = absoluteURL(quiz, self.request)
        return '%s/edit_grading_target.html?camefrom=%s' % (context_url,
                                                            camefrom)


class DeployedQuizStudentAnswersTableViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <tal:block i18n:domain="schooltool.quiz">
        <h3 i18n:translate="">Student Answers</h3>
        <p class="hint" i18n:translate="">
          Students of this section that have taken this quiz.
        </p>
        <div tal:define="ajax nocall:view/view/providers/ajax"
             tal:content="structure ajax/view/context/student_answers_table"
             />
      </tal:block>
    ''')


def student_answer_formatter(table):
    def formatter(value, item, formatter):
        deployed_url = absoluteURL(formatter.context, formatter.request)
        if table.show_names_button_id in formatter.request:
            fmt = '%s/answers.html?student=%s&show_student=true&camefrom=%s'
            url = fmt % (deployed_url, item.username, deployed_url)
        else:
            url = '%s/answers.html?student=%s&camefrom=%s' % (
                deployed_url, item.username, deployed_url)
            value = translate(_('Student'), context=formatter.request)
        if not formatter.context.open_by_teacher:
            return '<a href="%s">%s</a>' % (url, value)
        else:
            return value
    return formatter


def answer_grade_getter(person, formatter):
    return student_grade(formatter.context, person)


class StudentAnswersTable(table.ajax.Table):

    show_names_button_id = 'show-names-button'
    hide_names_button_id = 'hide-names-button'

    def items(self):
        result = []
        section = ISection(self.context)
        for student in section.members:
            if taken(student, self.context):
                result.append(student)
        return result

    def columns(self):
        student = table.column.LocaleAwareGetterColumn(
            name='student',
            title=_('Student'),
            getter=lambda i, f: i.title,
            cell_formatter=student_answer_formatter(self))
        grade = GetterColumn(
            name='grade',
            title=_('Grade'),
            getter=answer_grade_getter)
        return [student, grade]

    def sortOn(self):
        return (('student', False),)

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data student-answers-table'})


class StudentAnswersTableFilter(table.ajax.TableFilter):

    template = ViewPageTemplateFile('templates/answers_table_filter.pt')

    @property
    def label(self):
        if self.render_hide_names_button:
            return _('Hide Names')
        return _('Show Names')

    @property
    def render_show_names_button(self):
        return self.manager.show_names_button_id not in self.request

    @property
    def render_hide_names_button(self):
        return self.manager.show_names_button_id in self.request


class DeployedQuizViewDoneLinkViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
        <h3 i18n:domain="schooltool" class="done-link">
          <a tal:attributes="href view/url"
             i18n:translate="">
            Done
          </a>
        </h3>
    ''')

    def url(self):
        result = self.request.get('camefrom')
        if result is None:
            quiz = list(self.context.quizzes)[0]
            result = absoluteURL(quiz, self.request)
        return result


class DeployedQuizActionLinks(flourish.page.RefineLinksViewlet):

    pass


class DeployedQuizViewLinks(flourish.page.RefineLinksViewlet):

    pass


class DeployedQuizViewDeleteLinkViewlet(flourish.page.ModalFormLinkViewlet):

    @property
    def url(self):
        link = self.link
        if not link:
            return None
        url = "%s/%s" % (absoluteURL(self.context, self.request), link)
        camefrom = self.request.get('camefrom')
        if camefrom is not None:
            url += '?camefrom=%s' % camefrom
        return url

    @property
    def dialog_title(self):
        title = _('Delete this deployment?')
        return translate(title, context=self.request)


class DeployedQuizViewSolutionLinkViewlet(flourish.page.LinkViewlet):

    @property
    def url(self):
        url = super(DeployedQuizViewSolutionLinkViewlet, self).url
        camefrom = self.request.get('camefrom')
        if camefrom is not None:
            url += '?camefrom=%s' % camefrom
        return url


class DeployedQuizSolutionDoneLinkViewlet(SolutionDoneLinkViewlet):

    def url(self):
        url = absoluteURL(self.context, self.request)
        camefrom = self.request.get('camefrom')
        if camefrom is not None:
            url += '?camefrom=%s' % camefrom
        return url


class AnsweredEvaluation(Evaluation):

    def __init__(self, item, answer, grade):
        self.answer = answer
        evaluator = None
        super(AnsweredEvaluation, self).__init__(
            item, PercentScoreSystem, grade, evaluator)

class SavedAssignment(Assignment):

    def __init__(self, item, answer):
        self.answer = answer
        super(SavedAssignment, self).__init__(
            item)



class DeployedQuizTakeView(flourish.form.Form):

    implements(interfaces.ISchoolToolQuizView,
               interfaces.IMathJaxView)

    container_class = 'container widecontainer'
    id = 'take-quiz-form'
    content_template = ViewPageTemplateFile('templates/take_form.pt')
    legend = _('Questions')

    def __init__(self, *args, **kw):
        super(DeployedQuizTakeView, self).__init__(*args, **kw)
        self.values = {}
        if not self.context.open_by_teacher:
            person = IPerson(self.request.principal, None)
            if person is not None:
                url = '%s/quizzes.html' % absoluteURL(person, self.request)
            else:
                app = ISchoolToolApplication(None)
                url = absoluteURL(app, self.request)
            self.request.response.redirect(url)
        self.save_items()

    @property
    def title(self):
        return self.context.title

    @property
    def subtitle(self):
        section = ISection(self.context)
        return section.title

    @Lazy
    def types_vocabulary(self):
        factory = getUtility(IVocabularyFactory,
                             name='schooltool.quiz.quiz_item_types')
        return factory(None)

    @property
    def fields(self):
        form_fields = []
        for item in self.context:
            form_fields.append(
                self.createFormField(removeSecurityProxy(item)))
        if self.context.random:
            random.shuffle(form_fields)
        result = field.Fields()
        for form_field in form_fields:
            result += form_field
        return result
    
    @property
    def savetime(self):
        app = ISchoolToolApplication(None)
        preferences = interfaces.IApplicationPreferences(app)
        return int(preferences.save_time)*6000

    def createFormField(self, item):
        term = self.types_vocabulary.getTerm(item.type_)
        if term.token == 'open':
            return self.createOpenQuestionField(item)
        elif term.token == 'multiple':
            return self.createMultipleSelectionField(item)
        else:
            return self.createSelectionField(item)

    def createOpenQuestionField(self, item):
        name = item.__name__.encode('utf-8')
        person = IPerson(self.request.principal, None)
        saved_response = None
        assignments = None
        if person is not None:
            assignments = interfaces.IAssignments(removeSecurityProxy(person))
        if item.__name__ in assignments:
            if assignments[item.__name__].answer == None:
                saved_response = u''
            else:
                saved_response = unicode(assignments[item.__name__].answer)
        schema_field = interfaces.ReStructuredText(
            __name__=name,
            description=item.body,
            default=saved_response)
        self.values[name] = None
        datamanager.DictionaryField(self.values, schema_field)
        result = field.Fields(schema_field)
        return result

    def choicesVocabulary(self, item, shuffle=False):
        terms = []
        for index, choice in enumerate(item.choices):
            terms.append(SimpleTerm(token=index,
                                    value=choice,
                                    title=choice.body))
        if shuffle:
            random.shuffle(terms)
        return SimpleVocabulary(terms)

    def createMultipleSelectionField(self, item):
        name = item.__name__.encode('utf-8')
        choices_vocabulary = self.choicesVocabulary(item, self.context.random) 
        person = IPerson(self.request.principal, None)
        saved_response = None
        assignments = None
        if person is not None:
            assignments = interfaces.IAssignments(removeSecurityProxy(person))
        if item.__name__ in assignments:
            saved_response = assignments[item.__name__].answer
        schema_field = schema.List(
            __name__=name,
            description=item.body,
            value_type=schema.Choice(vocabulary=choices_vocabulary),
            min_length=1,
            default=saved_response,)
        self.values[name] = None
        datamanager.DictionaryField(self.values, schema_field)
        result = field.Fields(schema_field)
        result[name].widgetFactory = RestLabeledCheckBoxFieldWidget
        return result

    def createSelectionField(self, item):
        name = item.__name__.encode('utf-8')
        choices_vocabulary = self.choicesVocabulary(item, self.context.random) 
        person = IPerson(self.request.principal, None)
        saved_response = None
        assignments = None
        if person is not None:
            assignments = interfaces.IAssignments(removeSecurityProxy(person))
        if item.__name__ in assignments:
            saved_response = assignments[item.__name__].answer
        schema_field = schema.List(
            __name__=name,
            description=item.body,
            value_type=schema.Choice(vocabulary=choices_vocabulary),
            min_length=1,
            default = saved_response)
        self.values[name] = None
        datamanager.DictionaryField(self.values, schema_field)
        result = field.Fields(schema_field)
        result[name].widgetFactory = RestLabeledRadioFieldWidget
        return result

    def getContent(self):
        return self.values

    def save_items(self):
        person = IPerson(self.request.principal, None)
        if person is not None:
            try:
                data, errors = self.extractData()
            except AttributeError:
                data = {}
            assignments = interfaces.IAssignments(removeSecurityProxy(person))
            for item in self.context:
                item = removeSecurityProxy(item)
                name = item.__name__.encode('utf-8')
                if item.__name__ not in assignments: 
                    if name in data.keys():
                        answer = data[name]
                        assignment = SavedAssignment(item, answer)
                    else: 
                        assignment = SavedAssignment(item, None)
                    assignments.addAssignment(assignment)
                else:
                    if name in data.keys():
                        answer = data[name]
                    else: 
                        answer = None
                    if assignments[item.__name__].answer != answer and answer is not None:
                        assignments[item.__name__].answer = answer
        return


    @button.buttonAndHandler(_('Submit'), name='submit')
    def handle_submit(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        person = IPerson(self.request.principal, None)
        if person is not None:
            evaluations = IEvaluations(removeSecurityProxy(person))
            assignments = interfaces.IAssignments(removeSecurityProxy(person))
            for item in self.context:
                item = removeSecurityProxy(item)
                name = item.__name__.encode('utf-8')
                answer = data[name]
                term = self.types_vocabulary.getTerm(item.type_)
                if term.token == 'open':
                    grade = UNSCORED
                elif term.token == 'multiple':
                    grade = self.calculateMultipleSelectionGrade(item, answer)
                else:
                    grade = self.calculateSelectionGrade(item, answer)
                if item.__name__ in assignments:
                    del assignments[item.__name__]
                evaluation = AnsweredEvaluation(item, answer, grade)
                evaluations.addEvaluation(evaluation)
        self.request.response.redirect(self.nextURL())

    def calculateMultipleSelectionGrade(self, item, answer):
        choices_vocabulary = self.choicesVocabulary(item)
        given_choices = []
        for token in answer:
            given_choices.append(choices_vocabulary.getTerm(token).value)
        correct = []
        for choice in item.choices:
            if (choice.correct and choice in given_choices) or \
               (not choice.correct and choice not in given_choices):
                correct.append(choice)
        grade = len(correct) / float(len(item.choices))
        return grade * 100

    def calculateSelectionGrade(self, item, answer):
        choices_vocabulary = self.choicesVocabulary(item)
        given_choices = []
        for token in answer:
            given_choices.append(choices_vocabulary.getTerm(token).value)
        for choice in item.choices:
            if choice.correct and choice in given_choices:
                return 100
        return 0

    @button.buttonAndHandler(_('Cancel'), name='cancel')
    def handle_cancel(self, action):
        self.request.response.redirect(self.nextURL())

    def nextURL(self):
        person = IPerson(self.request.principal, None)
        if person is not None:
            return '%s/quizzes.html' % absoluteURL(person, self.request)
        section = ISection(self.context)
        return absoluteURL(section, self.request)

    def updateActions(self):
        super(DeployedQuizTakeView, self).updateActions()
        self.actions['submit'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def render_rst(self, rst):
        return render_rst(rst, self.request)


class DeployedQuizSaveView(flourish.page.Page):
    """ Quiz Auto-save handler. Ajax-called from QuizTakeView at 
        a global time interval defined in QuizPreferences"""    

    implements(interfaces.ISchoolToolQuizView)

    def __init__(self, *args, **kw):
        super(DeployedQuizSaveView, self).__init__(*args, **kw)
        person = IPerson(self.request.principal, None)
        
        if person is not None:
            data = self.request.form
            for key in data.keys():
                if key[-13:] == '-empty-marker':
                    del data[key]
                else:
                    newkey = key[13:]
                    data[newkey] = data.pop(key)
            assignments = interfaces.IAssignments(removeSecurityProxy(person))
            for item in self.context:
                item = removeSecurityProxy(item)
                name = item.__name__
                if name in data.keys():
                    answer = None
                    if item.type_ == u'Selection': 
                        answer = [item.choices[int(data[name])]]
                    elif item.type_ == u'Multiple Selection':
                        answer = [item.choices[int(data[name][k])] for k in range(0, len(data[name]))]
                    elif item.type_ == u'Open':
                        answer = str(data[name])
                    if assignments[item.__name__].answer != answer and answer is not None:
                        assignments[item.__name__].answer = answer


class DeployedQuizAnswersView(flourish.page.Page):

    implements(interfaces.ISchoolToolQuizView,
               interfaces.IMathJaxView)

    @Lazy
    def student(self):
        section = ISection(self.context)
        app = ISchoolToolApplication(None)
        persons = app['persons']
        student_id = self.request.get('student')
        student = persons.get(student_id)
        if student is None or \
           student not in section.members or \
           not flourish.canView(student):
            raise NotFound(self.context, student_id, self.request)
        return student

    def update(self):
        if self.context.open_by_teacher:
            person = IPerson(self.request.principal, None)
            if person is not None:
                url = '%s/quizzes.html' % absoluteURL(person, self.request)
            else:
                app = ISchoolToolApplication(None)
                url = absoluteURL(app, self.request)
            self.request.response.redirect(url)

    @property
    def tittle(self):
        return self.context.title

    @property
    def subtitle(self):
        if self.request.get('show_student'):
            return self.student.title


class DeployedQuizAnswersViewTableViewlet(flourish.viewlet.Viewlet):

    template = ViewPageTemplateFile('templates/answers.pt')

    def evaluations(self):
        person = self.view.student
        evaluations = IEvaluations(removeSecurityProxy(person))
        result = []
        for item in self.context:
            result.append(evaluations.get(item))
        return result

    def choicesVocabulary(self, item, shuffle=False):
        terms = []
        for index, choice in enumerate(item.choices):
            terms.append(SimpleTerm(token=index,
                                    value=choice,
                                    title=choice.body))
        if shuffle:
            random.shuffle(terms)
        return SimpleVocabulary(terms)

    @Lazy
    def types_vocabulary(self):
        factory = getUtility(IVocabularyFactory,
                             name='schooltool.quiz.quiz_item_types')
        return factory(None)

    def is_open(self, item):
        term = self.types_vocabulary.getTerm(item.type_)
        return term.token == 'open'

    def is_graded(self, evaluation):
        return evaluation and evaluation.value not in (UNSCORED, None)

    def is_correct(self, evaluation):
        return self.is_graded(evaluation) and evaluation.value >= 100

    def score_info(self, evaluation):
        if not self.is_graded(evaluation):
            return {
                'css': 'restructuredtext-widget not-graded',
                'label': _('NOT GRADED YET'),
                }
        if self.is_correct(evaluation):
            return {
                'css': 'restructuredtext-widget correct',
                'label': _('CORRECT'),
                }
        return {
            'css': 'restructuredtext-widget incorrect',
            'label': _('INCORRECT'),
            }

    def render_pre(self, text):
        return render_pre(text)

    def given_choices(self, item, answers):
        result = []
        for answer in answers:
            # XXX: this check is put here to avoid evolving the data
            if interfaces.IChoice.providedBy(answer):
                choice = answer
            else:
                # It's an old answer (only choice.body stored)
                choice = None
                choices_vocabulary = self.choicesVocabulary(item)
                for term in choices_vocabulary:
                    if term.title == answer:
                        choice = term.value
                        break
            if choice is not None:
                result.append(choice)
        return result

    def render_item(self, item, attr):
        renderer = flourish.form.Form(item, self.request)
        renderer.mode = DISPLAY_MODE
        schema_field = interfaces.IQuizItem[attr]
        renderer.fields = field.Fields(schema_field)
        renderer.update()
        return renderer.widgets[attr].render()

    def render_choice(self, choice, attr):
        renderer = flourish.form.Form(choice, self.request)
        renderer.mode = DISPLAY_MODE
        schema_field = interfaces.IChoice[attr]
        renderer.fields = field.Fields(schema_field)
        renderer.update()
        return renderer.widgets[attr].render()


class DeployedQuizAnswersViewDoneLinkViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
        <h3 i18n:domain="schooltool" class="done-link">
          <a tal:attributes="href view/url"
             i18n:translate="">
            Done
          </a>
        </h3>
    ''')

    def url(self):
        person = IPerson(self.request.principal)
        person_url = absoluteURL(person, self.request)
        url = self.request.get('camefrom',
                               '%s/quizzes.html' % person_url)
        return url


class DeployedQuizGradeOpenQuestionsLinkViewlet(flourish.page.LinkViewlet):

    @Lazy
    def types_vocabulary(self):
        factory = getUtility(IVocabularyFactory,
                             name='schooltool.quiz.quiz_item_types')
        return factory(None)

    @property
    def enabled(self):
        result = []
        for item in self.context:
            term = self.types_vocabulary.getTerm(item.type_)
            if term.token == 'open':
                result.append(item)
        return bool(result)


class DeployedQuizGradeOpenQuestionsView(flourish.form.Form):

    implements(interfaces.ISchoolToolQuizView,
               interfaces.IMathJaxView)

    container_class = 'container extra-wide-container'
    id = 'take-quiz-form'
    content_template = ViewPageTemplateFile('templates/take_form.pt')
    legend = _('Open Questions')

    def __init__(self, *args, **kw):
        super(DeployedQuizGradeOpenQuestionsView, self).__init__(*args, **kw)
        self.section = ISection(self.context)
        self.values = {}
        for item in self.open_questions:
            name = item.__name__.encode('utf-8')
            self.values[name] = []
            for student in self.section.members:
                evaluations = IEvaluations(removeSecurityProxy(student))
                evaluation = evaluations.get(item, None)
                if evaluation is not None:
                    if evaluation.value is not UNSCORED and evaluation.value:
                        self.values[name].append(student.username)

    def render_rst(self, rst):
        return render_rst(rst, self.request)

    @property
    def title(self):
        return self.context.title

    @Lazy
    def open_questions(self):
        result = []
        for item in self.context:
            term = self.types_vocabulary.getTerm(item.type_)
            if term.token == 'open':
                result.append(item)
        return result

    @property
    def savetime(self):
        return 1800000 #30 minutes

    @property
    def fields(self):
        show_names = 'show_names' in self.request
        result = field.Fields()
        for item in self.open_questions:
            item = removeSecurityProxy(item)
            name = item.__name__.encode('utf-8')
            answers = []
            for student in self.section.members:
                evaluations = IEvaluations(removeSecurityProxy(student))
                evaluation = evaluations.get(item, None)
                if evaluation is not None:
                    title = evaluation.answer
                    if show_names:
                        title = u'%s [%s]' % (title, student.title)
                    answers.append((student.username, title))
            answers_vocabulary = vocabulary(answers)
            schema_field = schema.List(
                __name__=name,
                description=item.body,
                value_type=schema.Choice(vocabulary=answers_vocabulary))
            datamanager.DictionaryField(self.values, schema_field)
            form_field = field.Fields(schema_field)
            form_field[name].widgetFactory = PreLabeledCheckBoxFieldWidget
            result += form_field
        return result

    def getContent(self):
        return self.values

    @Lazy
    def types_vocabulary(self):
        factory = getUtility(IVocabularyFactory,
                             name='schooltool.quiz.quiz_item_types')
        return factory(None)

    @button.buttonAndHandler(_('Submit'), name='submit')
    def handle_submit(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        for student in self.section.members:
            evaluations = IEvaluations(removeSecurityProxy(student))
            for item in self.open_questions:
                item = removeSecurityProxy(item)
                name = item.__name__.encode('utf-8')
                answers = data.get(name)
                if student.username in answers:
                    grade = 100.0
                else:
                    grade = 0
                evaluation = evaluations.get(item)
                if evaluation is not None:
                    new_evaluation = AnsweredEvaluation(item,
                                                        evaluation.answer,
                                                        grade)
                    new_evaluation.previous = evaluation
                    evaluations.addEvaluation(new_evaluation)
        self.request.response.redirect(self.nextURL())

    @button.buttonAndHandler(_('Show Names'), name='shownames')
    def handle_show_names(self, action):
        context_url = absoluteURL(self.context, self.request)
        url = '%s/grade_open_questions.html?show_names=true' % (
            context_url)
        self.request.response.redirect(url)

    @button.buttonAndHandler(_('Hide Names'), name='hidenames')
    def handle_hide_names(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    @button.buttonAndHandler(_('Cancel'), name='cancel')
    def handle_cancel(self, action):
        self.request.response.redirect(self.nextURL())

    def nextURL(self):
        return absoluteURL(self.context, self.request)

    def updateActions(self):
        super(DeployedQuizGradeOpenQuestionsView, self).updateActions()
        self.actions['submit'].addClass('button-ok')
        self.actions['hidenames'].addClass('button-cancel')
        self.actions['shownames'].addClass('button-cancel')
        self.actions['cancel'].addClass('button-cancel')
        if 'show_names' in self.request:
            del self.actions['shownames']
        else:
            del self.actions['hidenames']


class DeployedQuizFormScriptViewlet(flourish.viewlet.Viewlet):

    template = ViewPageTemplateFile('templates/deployed_form_script.pt')


class DeployedQuizGradingFormScriptViewlet(flourish.viewlet.Viewlet):

    template = ViewPageTemplateFile(
        'templates/deployed_grading_form_script.pt')


class UpdateGradesView(flourish.page.Page):

    def __call__(self):
        section = ISection(self.context)
        worksheet = None
        journal = None
        meeting = None
        if self.context.worksheet_id is not None:
            worksheets = IActivities(section)
            worksheet = worksheets[self.context.worksheet_id]
        elif self.context.period_id is not None:
            journal = ISectionJournal(section)
            meeting = journal.findMeeting(self.context.period_id)
        for student in section.members:
            student = removeSecurityProxy(student)
            grades = []
            for item in self.context:
                evaluations = IEvaluations(student)
                evaluation = evaluations.get(item)
                if evaluation is not None and \
                   evaluation.value is not UNSCORED:
                    grades.append(evaluation.value)
            if worksheet is not None:
                if grades:
                    grade = sum(grades) / float(len(grades))
                    grade = Decimal('%.1f' % grade)
                    gradebook = IGradebook(worksheet)
                    evaluator = None
                    activity = worksheet[self.context.activity_id]
                    gradebook.evaluate(student, activity, grade, evaluator)
            elif meeting is not None:
                if grades:
                    grade = sum(grades) / float(len(grades))
                    grade = '%.0f' % max(round(grade / 10.0), 1)
                    journal.setGrade(student, meeting, grade)
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)


class DeployedQuizTakeViewScriptViewlet(flourish.viewlet.Viewlet):

    template = ViewPageTemplateFile('templates/take_view_script.pt')

    container_id = 'timer-dialog-container'
    warning_times = (120,)  # seconds before end time

    @property
    def warnings(self):
        result = []
        end = self.context.end
        for warning_time in self.warning_times:
            interval = (end - self.now).seconds - warning_time
            if interval > 0:
                minutes = warning_time / 60
                message = translate(
                    _('You have ${minutes} minutes to submit this quiz',
                      mapping={'minutes': minutes}),
                    context=self.request)
                result.append({
                        'interval': interval,
                        'message': message,
                        })
        minutes = int(max(round((end - self.now).seconds / 60.0), 1))
        message = translate(
            _('You have ${minutes} minutes to submit this quiz',
              mapping={'minutes': minutes}),
            context=self.request)
        result.insert(0, {
                'interval': 0,
                'message': message,
                })
        return result

    @Lazy
    def now(self):
        return datetime.now()

    def dialog_title(self):
        return translate(_('Remaining time notification'),
                         context=self.request)


class OpenLinkViewlet(flourish.page.LinkViewlet):

    @property
    def title(self):
        if self.context.open_by_teacher:
            return _('Close')
        return _('Open')

    @property
    def url(self):
        url = super(OpenLinkViewlet, self).url
        camefrom = self.request.get('camefrom')
        if camefrom is not None:
            url += '?camefrom=%s' % camefrom
        return url


class DeployedQuizOpenView(flourish.page.Page):

    def __call__(self):
        current = self.context.open_by_teacher
        self.context.open_by_teacher = not current
        url = absoluteURL(self.context, self.request)
        camefrom = self.request.get('camefrom')
        if camefrom is not None:
            url += '?camefrom=%s' % camefrom
        self.request.response.redirect(url)
