#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation,
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""
SchoolTool Quiz quiz views.
"""

import re
import tempfile
import zipfile
from StringIO import StringIO
import xlwt
from operator import attrgetter

from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.component import adapts, getUtility
from zope.container.interfaces import INameChooser
from zope.dublincore.interfaces import IZopeDublinCore
from zope.event import notify
from zope.i18n import translate
from zope.interface import implements, directlyProvides
from zope.lifecycleevent import ObjectCreatedEvent
from zope.proxy import sameProxiedObjects
from zope.publisher.browser import BrowserView
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.schema.interfaces import IVocabularyFactory
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.traversing.browser.interfaces import IAbsoluteURL

from z3c.form import field, form, button
from z3c.form.interfaces import DISPLAY_MODE
from zc.table.column import GetterColumn
from zc.table.interfaces import ISortableColumn

from schooltool import table
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.basicperson.browser.person import EditPersonRelationships
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.export.export import Text, MegaExporter
from schooltool.export.importer import ImporterBase, FlourishMegaImporter
from schooltool.person.interfaces import IPerson
from schooltool.relationship.relationship import getRelatedObjects
from schooltool.requirement.interfaces import IEvaluations
from schooltool.skin import flourish

from schooltool.quiz import QuizMessage as _
from schooltool.quiz import interfaces
from schooltool.quiz import relationships
from schooltool.quiz.browser.widget import PersonSections
from schooltool.quiz.browser.widget import ReStructuredTextColumn
from schooltool.quiz.browser.widget import TypeColumn
from schooltool.quiz.browser.widget import editors_formatter
from schooltool.quiz.browser.widget import ImageDataConverter
from schooltool.quiz.quiz import Quiz
from schooltool.quiz.quizitem import QuizItem, Choice
from schooltool.quiz.quizimage import QuizImage
from schooltool.quiz.utils import render_rst, render_date


choices_row_separator = '----------\n'
choices_column_separator = '**|**'
QUIZ_JSON_NAME = 'quiz.json'


class QuizContainerAbsoluteURLAdapter(BrowserView):

    adapts(interfaces.IQuizContainer, IBrowserRequest)
    implements(IAbsoluteURL)

    def __str__(self):
        app = ISchoolToolApplication(None)
        return '%s/quizzes' % absoluteURL(app, self.request)

    __call__ = __str__


class QuizzesLink(flourish.page.LinkViewlet, PersonSections):

    startup_view_name = 'quizzes.html'

    @Lazy
    def person(self):
        return IPerson(self.request.principal, None)

    @property
    def enabled(self):
        if self.person is None:
            return False
        editable_quizzes = getRelatedObjects(
            self.person, relationships.URIQuiz, relationships.URIQuizEditors)
        return self.instructor_sections or \
               self.learner_sections or \
               editable_quizzes

    @property
    def url(self):
        if self.person is None:
            return ''
        return '%s/%s' % (absoluteURL(self.person, self.request),
                          self.startup_view_name)


def taken(person, deployed):
    result = False
    evaluations = IEvaluations(removeSecurityProxy(person))
    for item in deployed:
        evaluation = evaluations.get(item, None)
        if evaluation is not None:
            return True
    return result


class PersonQuizzesView(flourish.page.Page, PersonSections):

    implements(interfaces.ISchoolToolQuizView,
               interfaces.IPersonQuizView)

    container_class = 'container widecontainer'

    @Lazy
    def person(self):
        return self.context

    @property
    def quizContainer(self):
        app = ISchoolToolApplication(None)
        return interfaces.IQuizContainer(app)

    @Lazy
    def editable_quizzes(self):
        result = []
        for quiz in self.quizContainer.values():
            if self.person in quiz.editors:
                result.append(quiz)
        return result

    @Lazy
    def deployed_quizzes(self):
        result = []
        for section in self.instructor_sections:
            container = interfaces.IDeployedQuizContainer(section)
            for deployed in container.values():
                result.append(deployed)
        return result

    @Lazy
    def takeable_quizzes(self):
        result = []
        for section in self.learner_sections:
            container = interfaces.IDeployedQuizContainer(section)
            for deployed in container.values():
                if deployed.open_by_teacher and not self.taken(deployed):
                    result.append(deployed)
        return result

    @Lazy
    def taken_quizzes(self):
        result = []
        for section in self.learner_sections:
            container = interfaces.IDeployedQuizContainer(section)
            for deployed in container.values():
                if self.taken(deployed):
                    result.append(deployed)
        return result

    def taken(self, deployed):
        return taken(self.context, deployed)


class QuizContainerView(flourish.page.Page):

    def update(self):
        print list(self.context.keys())


class QuizContainerAddLinks(flourish.page.RefineLinksViewlet):

    pass


class QuizContainerActionsLinks(flourish.page.RefineLinksViewlet):

    pass


class QuizAddLinks(flourish.page.RefineLinksViewlet):

    pass


class QuizViewLinks(flourish.page.RefineLinksViewlet):

    pass


class QuizActionLinks(flourish.page.RefineLinksViewlet):

    pass


class QuizAddLink(flourish.page.LinkViewlet):

    @property
    def url(self):
        app = ISchoolToolApplication(None)
        container = interfaces.IQuizContainer(app)
        return '%s/add.html' % absoluteURL(container, self.request)


class QuizDeployLinkViewlet(flourish.page.LinkViewlet):

    @property
    def enabled(self):
        return len(list(self.context))


class PersonQuizAddLink(QuizAddLink, PersonSections):

    @Lazy
    def person(self):
        return IPerson(self.request.principal, None)

    @property
    def enabled(self):
        if self.person is not None:
            return self.instructor_sections
        return False

    @property
    def url(self):
        app = ISchoolToolApplication(None)
        container = interfaces.IQuizContainer(app)
        camefrom = '%s/quizzes.html' % absoluteURL(self.context, self.request)
        return '%s/add.html?camefrom=%s' % (
            absoluteURL(container, self.request), camefrom)


class PersonQuizImportLink(QuizAddLink, PersonSections):

    @Lazy
    def person(self):
        return IPerson(self.request.principal, None)

    @property
    def enabled(self):
        if self.person is not None:
            return self.instructor_sections
        return False

    @property
    def url(self):
        app = ISchoolToolApplication(None)
        container = interfaces.IQuizContainer(app)
        camefrom = '%s/quizzes.html' % absoluteURL(self.context, self.request)
        return '%s/import.html?camefrom=%s' % (
            absoluteURL(container, self.request), camefrom)


class QuizSolutionLinkViewlet(flourish.page.LinkViewlet):

    @property
    def enabled(self):
        return len(list(self.context))


class QuizView(flourish.page.Page):

    implements(interfaces.ISchoolToolQuizView,
               interfaces.IMathJaxView)

    container_class = 'container widecontainer'

    @property
    def title(self):
        return self.context.title

    @property
    def can_modify(self):
        person = IPerson(self.request.principal, None)
        return person in self.context.editors

    @Lazy
    def quiz_items(self):
        return list(self.context)

    @Lazy
    def deployed_quizzes(self):
        return list(self.context.deployed)


class EditableQuizzesViewlet(flourish.viewlet.Viewlet, PersonSections):

    template = InlineViewPageTemplate('''
      <tal:block i18n:domain="schooltool.quiz"
                 tal:define="quizzes view/view/editable_quizzes">
        <h3 i18n:translate="">Editable Quizzes</h3>
        <div tal:condition="quizzes"
             tal:define="ajax nocall:view/view/providers/ajax"
             tal:content="structure ajax/view/context/editable_quizzes_table"
             />
        <p i18n:translate="" tal:condition="not:quizzes">
          There are no quizzes you can edit yet.
        </p>
      </tal:block>
    ''')

    @Lazy
    def person(self):
        return self.context

    @property
    def enabled(self):
        editable_quizzes = getRelatedObjects(self.person,
                                             relationships.URIQuiz,
                                             relationships.URIQuizEditors)
        return self.instructor_sections or editable_quizzes

    def render(self, *args, **kw):
        if self.enabled:
            return self.template(*args, **kw)


class EditableQuizzesTable(table.ajax.Table):

    def items(self):
        return self.view.editable_quizzes

    def columns(self):
        default = super(EditableQuizzesTable, self).columns()
        questions = GetterColumn(
            name='questions',
            title=_('Questions'),
            getter=lambda quiz, formatter: len(list(quiz)))
        editors = GetterColumn(
            name='editors',
            title=_('Editors'),
            getter=lambda quiz, formatter: quiz.editors,
            cell_formatter=editors_formatter)
        created = GetterColumn(
            name='created',
            title=_('Created'),
            getter=lambda i, f: IZopeDublinCore(i).created,
            cell_formatter=lambda v, i, f: render_date(v),
            subsort=True)
        directlyProvides(created, ISortableColumn)
        duplicate = QuizActionColumn(
            'duplicate',
            title=_('Duplicate this quiz'),
            action='duplicate_quiz.html',
            library='schooltool.quiz.flourish',
            image='duplicate-icon.png')
        return default + [questions, editors, created, duplicate]

    def sortOn(self):
        return (('created', True),)

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data editable-quizzes-table'})


class QuizAddView(flourish.form.AddForm):

    legend = _('Quiz Information')
    fields = field.Fields(interfaces.IQuiz['title'])
    content_template = ViewPageTemplateFile('templates/form.pt')

    def updateActions(self):
        super(QuizAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def create(self, data):
        quiz = Quiz()
        form.applyChanges(self, quiz, data)
        return quiz

    def add(self, quiz):
        chooser = INameChooser(self.context)
        name = chooser.chooseName('', quiz)
        self.context[name] = quiz
        self.setCreator(quiz)
        self._quiz = quiz
        return quiz

    def nextURL(self):
        if self._finishedAdd:
            url = absoluteURL(self._quiz, self.request)
        else:
            url = self.request.get('camefrom',
                                   absoluteURL(self.context, self.request))
        return url

    def setCreator(self, quiz):
        creator = IPerson(self.request.principal)
        quiz.editors.add(removeSecurityProxy(creator))


class QuizEditView(flourish.form.Form, form.EditForm):

    legend = _('Quiz Information')
    fields = field.Fields(interfaces.IQuiz['title'])
    content_template = ViewPageTemplateFile('templates/form.pt')

    @property
    def title(self):
        return self.context.title

    def update(self):
        return form.EditForm.update(self)

    def updateActions(self):
        super(QuizEditView, self).updateActions()
        self.actions['submit'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def nextURL(self):
        return absoluteURL(self.context, self.request)

    @button.buttonAndHandler(_('Submit'), name='submit')
    def handle_submit(self, action):
        super(QuizEditView, self).handleApply.func(self, action)
        if (self.status == self.successMessage or
            self.status == self.noChangesMessage):
            self.request.response.redirect(self.nextURL())

    @button.buttonAndHandler(_('Cancel'), name='cancel')
    def handle_cancel(self, action):
        self.request.response.redirect(self.nextURL())


class QuizEditorsView(EditPersonRelationships):

    current_title = _('Current editors')
    available_title = _('Add editors')

    @property
    def title(self):
        return self.context.title

    def getCollection(self):
        return self.context.editors


class QuizDeleteView(flourish.form.DialogForm):

    template = ViewPageTemplateFile('templates/confirm_delete_quiz.pt')

    dialog_submit_actions = ('delete',)
    dialog_close_actions = ('cancel',)
    label = None

    @button.buttonAndHandler(_('Delete'), name='delete')
    def handle_delete(self, action):
        url = '%s/delete.html?delete.%s&CONFIRM' % (
            absoluteURL(self.context.__parent__, self.request),
            self.context.__name__)
        self.request.response.redirect(url)
        self.ajax_settings['dialog'] = 'close'

    @button.buttonAndHandler(_('Cancel'), name='cancel')
    def handle_cancel(self, action):
        pass

    def updateActions(self):
        super(QuizDeleteView, self).updateActions()
        self.actions['delete'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class QuizDuplicateView(flourish.page.Page):

    def __call__(self):
        container = self.context.__parent__
        copy = removeSecurityProxy(self.context).copy()
        chooser = INameChooser(container)
        name = chooser.chooseName('', copy)
        container[name] = copy
        copy.title = translate(
            _('DUPLICATED: ${quiz}', mapping={'quiz': self.context.title}),
            context=self.request)
        person = IPerson(self.request.principal, None)
        if person is not None:
            copy.editors.add(removeSecurityProxy(person))
        notify(ObjectCreatedEvent(copy))
        camefrom = self.request.get('camefrom',
                                    absoluteURL(container, self.request))
        self.request.response.redirect(camefrom)


class QuizContainerDeleteView(flourish.containers.ContainerDeleteView):

    def nextURL(self):
        if 'CONFIRM' in self.request:
            person = IPerson(self.request.principal)
            url = '%s/quizzes.html' % absoluteURL(person, self.request)
            return url
        return super(QuizContainerDeleteView, self).nextURL()


class QuizSolutionView(flourish.page.Page):

    implements(interfaces.ISchoolToolQuizView,
               interfaces.IMathJaxView)

    container_class = 'container extra-wide-container'

    @property
    def title(self):
        return self.context.title

    @Lazy
    def quiz_items(self):
        return self.context


class QuizTitleViewlet(flourish.viewlet.Viewlet):

    template = ViewPageTemplateFile('templates/quiz_title.pt')


class QuizItemsViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <tal:block i18n:domain="schooltool.quiz"
                 tal:define="items view/view/quiz_items">
        <h3 i18n:translate="">Questions</h3>
        <div tal:condition="items"
             tal:define="ajax nocall:view/view/providers/ajax"
             tal:content="structure ajax/view/context/quiz_items_table"
             />
        <p i18n:translate="" tal:condition="not:items">
          This quiz has no questions yet.
        </p>
      </tal:block>
    ''')


class ActionColumn(table.column.ImageInputColumn):

    def __init__(self, name, title=None, action=None, library=None,
                 image=None):
        super(ActionColumn, self).__init__(
            'item.', '', name, library=library, image=image)
        self.alt = title
        self.link_title = title
        self.action = action

    def template(self):
        return '\n'.join([
                '<a href="%(href)s" title="%(title)s">',
                '<img src="%(src)s" alt="%(alt)s" />',
                '</a>'
                ])


class QuizActionColumn(ActionColumn):

    def params(self, item, formatter):
        result = super(QuizActionColumn, self).params(item, formatter)
        person_url = absoluteURL(formatter.context, formatter.request)
        camefrom = '%s/quizzes.html' % person_url
        quiz_url = absoluteURL(item, formatter.request)
        result['title'] = translate(self.link_title,
                                    context=formatter.request) or ''
        result['href'] = '%s/%s?camefrom=%s' % (
            quiz_url, self.action, camefrom)
        return result


class ItemActionColumn(ActionColumn):

    def params(self, item, formatter):
        result = super(ItemActionColumn, self).params(item, formatter)
        quiz_url = absoluteURL(formatter.context, formatter.request)
        result['title'] = translate(self.link_title,
                                    context=formatter.request) or ''
        result['href'] = '%s/%s?item_id=%s&camefrom=%s' % (
            quiz_url, self.action, item.__name__, quiz_url)
        return result


class ModalItemActionColumn(ItemActionColumn):

    template = ViewPageTemplateFile('templates/modal_item_column.pt')

    def renderCell(self, item, formatter):
        params = self.params(item, formatter)
        params['dialog_title'] = translate(_('Remove this question?'),
                                           context=formatter.request)
        self.context = item
        self.request = formatter.request
        return self.template(params=params)


class PositionColumn(GetterColumn):

    def getter(self, item, formatter):
        for position, current in enumerate(formatter.context):
            if sameProxiedObjects(current, item):
                return position

    def cell_formatter(self, value, item, formatter):
        template = '<a href="%(href)s">%(title)s</a>'
        quiz_url = absoluteURL(formatter.context, formatter.request)
        href = '%s/view_item.html?item_id=%s' % (quiz_url, item.__name__)
        params = {'href': href, 'title': value + 1}
        return template % params


class SortColumn(PositionColumn):

    def cell_formatter(self, value, item, formatter):
        template = '<select name="%(name)s" class="item-sort">%(options)s</select>'
        options = []
        for position, current in enumerate(formatter.context):
            display_position = position + 1
            option = '<option value="%d">%d</option>'
            if sameProxiedObjects(current, item):
                option = '<option value="%d" selected="selected">%d</option>'
            options.append(option % (position, display_position))
        params = {
            'name': item.__name__,
            'options': ''.join(options),
            }
        return template % params


class QuizItemsTable(table.ajax.Table):

    batch_size = 0

    def columns(self):
        position = PositionColumn(
            name='position',
            title=_('No.'))
        body = ReStructuredTextColumn(
            name='body',
            title=_('Body'),
            schema_field=interfaces.IQuizItem['body']
            )
        type_ = TypeColumn(
            name='type',
            title=_('Type'))
        result = [position, body, type_]
        result.extend(self.getActionColumns())
        return result

    def getActionColumns(self):
        # XXX: set action columns only if principal is editor
        sort = SortColumn(
            name='sort',
            title=u'')
        edit = ItemActionColumn(
            'edit',
            title=_('Edit this item'),
            action='edit_item.html',
            library='schooltool.skin.flourish',
            image='edit-icon.png')
        duplicate = ItemActionColumn(
            'duplicate',
            title=_('Duplicate this item'),
            action='duplicate_item.html',
            library='schooltool.quiz.flourish',
            image='duplicate-icon.png')
        remove = ModalItemActionColumn(
            'remove',
            title=_('Remove this item'),
            action='remove_item.html',
            library='schooltool.skin.flourish',
            image='remove-icon.png')
        return [sort, edit, duplicate, remove]

    def items(self):
        return self.view.quiz_items

    def sortOn(self):
        return (('position', False),)

    def updateFormatter(self):
        # XXX: set -with-actions class only if principal is editor
        klass = 'data quiz-items-table quiz-items-table-with-actions'
        css_classes = {'table': klass}
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes=css_classes)

    def update(self):
        changePosition = self.request.get('changePosition')
        if changePosition is not None:
            toChange = self.request.get(changePosition)
            if toChange is not None:
                new_position = int(toChange)
                self.context.updateOrder(changePosition, new_position)
        super(QuizItemsTable, self).update()


class QuizViewDoneLinkViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
        <h3 i18n:domain="schooltool" class="done-link">
          <a tal:attributes="href view/url"
             i18n:translate="">
            Done
          </a>
        </h3>
    ''')

    def url(self):
        person = IPerson(self.request.principal, None)
        if person is not None:
            return '%s/quizzes.html' % (absoluteURL(person, self.request))


class QuizDeleteLinkViewlet(flourish.page.ModalFormLinkViewlet):

    @property
    def dialog_title(self):
        title = _('Delete this quiz?')
        return translate(title, context=self.request)


class QuizSolutionTableViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <div tal:define="ajax nocall:view/view/providers/ajax"
           tal:content="structure ajax/view/context/solution_table" />
    ''')


class SolutionPositionColumn(PositionColumn):

    def cell_formatter(self, value, item, formatter):
        return value + 1


def solution_question_formatter(value, item, formatter):
    return render_rst(item.body, formatter.request)


def solution_solution_formatter(value, item, formatter):
    if item.type_ == u'Open':
        return render_rst(item.solution, formatter.request)
    choices = []
    for choice in item.choices:
        body = render_rst(choice.body, formatter.request)
        if choice.correct:
            choices.append(
                '<span class="correct">%s</span>' % body)
        else:
            choices.append(body)
    return ''.join(choices)


class SolutionTable(table.ajax.Table):

    def items(self):
        return self.view.quiz_items

    def columns(self):
        position = SolutionPositionColumn(
            name='position',
            title=_('No.'))
        type_ = TypeColumn(
            name='type',
            title=_('Type'))
        question = GetterColumn(
            name='question',
            title=_('Question'),
            getter=lambda i, f: '',
            cell_formatter=solution_question_formatter)
        solution = GetterColumn(
            name='solution',
            title=_('Solution'),
            getter=lambda i, f: '',
            cell_formatter=solution_solution_formatter)
        return [position, type_, question, solution]

    def sortOn(self):
        return (('position', False),)

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data solution-table'})


class SolutionDoneLinkViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <h3 i18n:domain="schooltool" class="done-link">
        <a tal:attributes="href view/url" i18n:translate="">Done</a>
      </h3>
    ''')

    def url(self):
        default = absoluteURL(self.context, self.request)
        return self.request.get('camefrom', default)


class QuizImportView(flourish.page.Page):

    hasErrors = False
    success = False
    data_provided = False
    content_template = ViewPageTemplateFile('templates/quiz_import.pt')

    @Lazy
    def quiz_container(self):
        app = ISchoolToolApplication(None)
        return interfaces.IQuizContainer(app)

    @Lazy
    def quiz_item_container(self):
        app = ISchoolToolApplication(None)
        return interfaces.IQuizItemContainer(app)

    @Lazy
    def quiz_image_container(self):
        app = ISchoolToolApplication(None)
        return interfaces.IQuizImageContainer(app)

    def import_quiz_data(self, quiz_info, z):
        factory = getUtility(IVocabularyFactory,
                             name='schooltool.quiz.quiz_item_types')
        vocabulary = factory(None)
        result = Quiz()
        result.title = translate(
            _('IMPORTED: ${title}', mapping={'title': quiz_info['title']}),
            context=self.request)
        quiz_container_chooser = INameChooser(self.quiz_container)
        name = quiz_container_chooser.chooseName('', result)
        self.quiz_container[name] = result
        self.setCreator(result)
        quiz_item_container_chooser = INameChooser(self.quiz_item_container)
        quiz_image_container_chooser = INameChooser(self.quiz_image_container)
        for item_info in quiz_info['items']:
            image_names = {}
            item = QuizItem()
            item.type_ = vocabulary.getTermByToken(item_info['type']).value
            # XXX: check for course/skill
            choices = []
            for choice_info in item_info['choices']:
                choice = Choice()
                choice.body = choice_info['body']
                choice.correct = choice_info['correct']
                choices.append(choice)
            item.choices = choices
            item_name = quiz_item_container_chooser.chooseName('', item)
            self.quiz_item_container[item_name] = item
            self.setQuiz(item, result)
            notify(ObjectCreatedEvent(item))
            for image_name, image_title in item_info['images'].items():
                try:
                    info = z.getinfo(image_name)
                except (KeyError,):
                    info = None
                if info is None:
                    continue
                contents = StringIO(z.read(info))
                field = interfaces.IQuizImage['contents']
                converter = ImageDataConverter(field, None)
                image = QuizImage()
                image.title = image_title
                image.contents = converter.toFieldValue(contents)
                new_image_name = quiz_image_container_chooser.chooseName('', image)
                self.quiz_image_container[new_image_name] = image
                image_names[image_name] = new_image_name
                notify(ObjectCreatedEvent(image))
                image.items.add(item)
            item.body = self.fix_image_names(item_info['body'], image_names)
            item.solution = self.fix_image_names(item_info['solution'], image_names)
            for choice in item.choices:
                choice.body = self.fix_image_names(choice.body, image_names)
        notify(ObjectCreatedEvent(result))

    def setQuiz(self, item, quiz):
        quiz.order.append(item.__name__)
        item.quizzes.add(quiz)

    def fix_image_names(self, text, image_names):
        for old_name, new_name in image_names.items():
            text = text.replace('/%s/inline' % old_name, '/%s/inline' % new_name)
        return text

    def setCreator(self, quiz):
        creator = IPerson(self.request.principal)
        quiz.editors.add(removeSecurityProxy(creator))

    def import_quiz(self, z):
        decoder = flourish.tal.JSONDecoder()
        try:
            info = z.getinfo(QUIZ_JSON_NAME)
        except (KeyError,):
            return
        contents = z.read(info)
        quiz_info = decoder.decode(contents)
        self.import_quiz_data(quiz_info, z)

    def update(self):
        if "UPDATE_SUBMIT" not in self.request:
            return
        zip_file = self.request.get('zipfile', '')
        if not zip_file:
            return
        try:
            with zipfile.ZipFile(zip_file, 'r') as z:
                self.import_quiz(z)
        except (zipfile.BadZipfile,):
            self.data_provided = True
            self.textareaErrors = _("Couldn't import quiz")
            return
        self.request.response.redirect(self.nextURL())

    def nextURL(self):
        return self.request.get('camefrom',
                                absoluteURL(self.context, self.request))


class QuizExportView(flourish.page.Page):

    MAX_SIZE = 1024 * (10 * 1024)

    def setUpHeaders(self):
        response = self.request.response
        response.setHeader('Content-Type', 'application/x-zip-compressed')

    def __call__(self):
        with tempfile.SpooledTemporaryFile(max_size=self.MAX_SIZE) as tmp:
            with zipfile.ZipFile(tmp, 'w', zipfile.ZIP_DEFLATED) as z:
                self.export_quiz(z)
            tmp.seek(0)
            self.setUpHeaders()
            return tmp.read()

    def export_quiz(self, z):
        factory = getUtility(IVocabularyFactory,
                             name='schooltool.quiz.quiz_item_types')
        vocabulary = factory(None)
        encoder = flourish.tal.JSONEncoder()
        result = {
            'title': self.context.title,
            'items': [],
            }
        for item in self.context:
            result['items'].append({
                    'body': item.body,
                    'type': vocabulary.getTerm(item.type_).token,
                    'choices': self.get_choices(item),
                    'solution': item.solution or '',
                    'images': self.get_images(item, z),
                    'course': getattr(item.course, 'course_id', ''),
                    'skill': self.get_skill(item),
                    })
        z.writestr('quiz.json', encoder.encode(result))

    def get_skill(self, item):
        result = {}
        if item.skill is not None:
            skill = removeSecurityProxy(item.skill)
            result['skill_id'] = skill.__name__
            result['skillset_id'] = skill.__parent__.__name__
        return result

    def get_choices(self, item):
        result = []
        if item.choices is not None:
            for choice in item.choices:
                result.append({
                        'body': choice.body,
                        'correct': choice.correct,
                        })
        return result

    @Lazy
    def image_container(self):
        app = ISchoolToolApplication(None)
        return interfaces.IQuizImageContainer(app)

    def get_images(self, item, z):
        result = {}
        text = '\n'.join([
                item.body,
                item.solution or '',
                '\n'.join([c.body for c in (item.choices or ())]),
                ])
        images = re.findall(r'quiz_images/(.*?)/inline', text.strip())
        for name in images:
            try:
                info = z.getinfo(name)
            except (KeyError,):
                info = None
            if info is not None:
                continue
            image = self.image_container.get(name)
            if image is None:
                continue
            contents = removeSecurityProxy(image.contents).openDetached()
            z.writestr(name, contents.read())
            contents.close()
            result[name] = image.title
        return result


class QuizHelpLinks(flourish.page.RefineLinksViewlet, PersonSections):

    @Lazy
    def person(self):
        return IPerson(self.request.principal, None)

    @property
    def enabled(self):
        if self.person is not None:
            editable_quizzes = getRelatedObjects(self.person,
                                                 relationships.URIQuiz,
                                                 relationships.URIQuizEditors)
            return self.instructor_sections or editable_quizzes

    def render(self, *args, **kw):
        if self.enabled:
            return super(QuizHelpLinks, self).render(*args, **kw)
        return ''


class QuizEditingHelpViewlet(flourish.page.LinkIdViewlet):

    template = InlineViewPageTemplate('''
    <a tal:attributes="href view/url;
                       onclick view/script"
       tal:content="view/title" />
    ''')

    @property
    def enabled(self):
        app = ISchoolToolApplication(None)
        preferences = interfaces.IApplicationPreferences(app)
        return preferences.mathjax_help is not None

    @property
    def script(self):
        url = self.url
        name = self.html_id
        params = {
            'width': 640,
            'height': 480,
            'resizable': 'yes',
            'scrollbars': 'yes',
            'toolbar': 'no',
            'location': 'no',
            }
        return "window.open('%s', '%s', '%s'); return false" % (
            url, name, ','.join(['%s=%s' % (k, v) for k, v in params.items()]))


class QuizEditingHelpView(flourish.page.Page):

    implements(interfaces.ISchoolToolQuizView,
               interfaces.IMathJaxView)

    template = InlineViewPageTemplate('''
    <tal:block content="structure view/mathjax_help" />
    ''')

    def mathjax_help(self):
        app = ISchoolToolApplication(None)
        preferences = interfaces.IApplicationPreferences(app)
        return self.render_rst(preferences,
                               interfaces.IApplicationPreferences,
                               'mathjax_help')

    def render_rst(self, value, iface, field_name):
        renderer = flourish.form.Form(value, self.request)
        renderer.mode = DISPLAY_MODE
        renderer.fields = field.Fields(iface[field_name])
        renderer.update()
        return renderer.widgets[field_name].render()


class UpdateMathJax(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <script>
        MathJax.Hub.Queue(['Typeset', MathJax.Hub]);
      </script>
    ''')
