#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2013 Shuttleworth Foundation,
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""
SchoolTool Quiz image views.
"""

from cStringIO import StringIO
from mimetypes import guess_all_extensions
from urllib import urlopen

try:
    import Image
except ImportError:
    from PIL import Image

from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.component import adapts
from zope.container.interfaces import INameChooser
from zope.dublincore.interfaces import IZopeDublinCore
from zope.event import notify
from zope.i18n import translate
from zope.interface import implements, directlyProvides
from zope.lifecycleevent import ObjectCreatedEvent
from zope.publisher.browser import BrowserView
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.publisher.interfaces import NotFound
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.traversing.browser.interfaces import IAbsoluteURL

from z3c.form import field, form, button
from z3c.form.interfaces import DISPLAY_MODE
from zc.table.column import GetterColumn
from zc.table.interfaces import ISortableColumn

from schooltool import table
from schooltool.app.browser.app import EditRelationships
from schooltool.app.browser.app import RelationshipAddTableMixin
from schooltool.app.browser.app import RelationshipRemoveTableMixin
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.common.fields import ImageFile
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.person.interfaces import IPerson
from schooltool.skin import flourish

from schooltool.quiz import QuizMessage as _
from schooltool.quiz import interfaces
from schooltool.quiz.browser.widget import PersonSections
from schooltool.quiz.browser.widget import ReStructuredTextColumn
from schooltool.quiz.browser.widget import TypeColumn
from schooltool.quiz.quizimage import QuizImage
from schooltool.quiz.utils import render_date, APP_ROOT_PATTERN


class QuizImageContainerAbsoluteURLAdapter(BrowserView):

    adapts(interfaces.IQuizImageContainer, IBrowserRequest)
    implements(IAbsoluteURL)

    def __str__(self):
        app = ISchoolToolApplication(None)
        return '%s/quiz_images' % absoluteURL(app, self.request)

    __call__ = __str__


class QuizImageContainerView(flourish.page.Page):

    container_class = 'container widecontainer'


class QuizImagesTableViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <div tal:define="ajax nocall:view/view/providers/ajax"
           tal:content="structure ajax/view/context/quiz_images_table"
           />
    ''')


class QuizImagesTable(table.ajax.Table):

    def columns(self):
        default = super(QuizImagesTable, self).columns()
        created = GetterColumn(
            name='created',
            title=_('Created'),
            getter=lambda i, f: IZopeDublinCore(i).created,
            cell_formatter=lambda v, i, f: render_date(v),
            subsort=True)
        directlyProvides(created, ISortableColumn)
        return default + [created]

    def sortOn(self):
        return (('created', True),)

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data quiz-images-table'})


class QuizImageAddView(flourish.form.AddForm):

    legend = _('Quiz Image Information')
    fields = field.Fields(interfaces.IQuizImage).select('title', 'contents')
    content_template = ViewPageTemplateFile('templates/form.pt')

    def updateActions(self):
        super(QuizImageAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def create(self, data):
        image = QuizImage()
        form.applyChanges(self, image, data)
        return image

    def add(self, image):
        chooser = INameChooser(self.context)
        name = chooser.chooseName('', image)
        self.context[name] = image
        self.setCreator(image)
        self._image = image
        return image

    def nextURL(self):
        if self._finishedAdd:
            url = absoluteURL(self._image, self.request)
        else:
            url = self.request.get('camefrom',
                                   absoluteURL(self.context, self.request))
        return url

    def setCreator(self, image):
        creator = IPerson(self.request.principal)
        image.editors.add(removeSecurityProxy(creator))


class QuizImageContainerAddLinks(flourish.page.RefineLinksViewlet):

    pass


class QuizImageAddLink(flourish.page.LinkViewlet):

    @property
    def url(self):
        app = ISchoolToolApplication(None)
        container = interfaces.IQuizImageContainer(app)
        return '%s/add.html' % absoluteURL(container, self.request)


class QuizImageView(flourish.page.Page):

    implements(interfaces.ISchoolToolQuizView)

    container_class = 'container widecontainer'

    @property
    def subtitle(self):
        return self.context.title


class QuizImageInlineView(flourish.widgets.ImageView):

    attribute = 'contents'


class QuizImageDownloadView(QuizImageInlineView):

    def renderImage(self, image):
        result = super(QuizImageDownloadView, self).renderImage(image)
        self.setDispositionHeader(image)
        return result

    def setDispositionHeader(self, image):
        extension = guess_all_extensions(image.mimeType)[-1]
        filename = self.context.__name__
        if extension:
            filename = '%s%s' % (filename, extension)
        disposition = 'attachment; filename="%s"' % filename
        self.request.response.setHeader('Content-Disposition', disposition)


class QuizImageEditView(flourish.form.Form, form.EditForm):

    legend = _('Quiz Image Information')
    content_template = ViewPageTemplateFile('templates/form.pt')
    fields = field.Fields(interfaces.IQuizImage).select('title')

    @property
    def title(self):
        return self.context.title

    def update(self):
        return form.EditForm.update(self)

    def updateActions(self):
        super(QuizImageEditView, self).updateActions()
        self.actions['submit'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def nextURL(self):
        camefrom = self.request.get('camefrom')
        url = absoluteURL(self.context, self.request)
        if camefrom is not None:
            return '%s?camefrom=%s' % (url, camefrom)
        return url

    @button.buttonAndHandler(_('Submit'), name='submit')
    def handle_submit(self, action):
        super(QuizImageEditView, self).handleApply.func(self, action)
        if (self.status == self.successMessage or
            self.status == self.noChangesMessage):
            self.request.response.redirect(self.nextURL())

    @button.buttonAndHandler(_('Cancel'), name='cancel')
    def handle_cancel(self, action):
        self.request.response.redirect(self.nextURL())


class QuizImageActionLinks(flourish.page.RefineLinksViewlet):

    pass


class QuizImageInsertLinks(flourish.page.RefineLinksViewlet):

    pass


class QuizItemAddViewQuizImageInsertLinks(flourish.page.RefineLinksViewlet):

    pass


class QuizImageDetailsViewlet(flourish.form.FormViewlet):

    template = ViewPageTemplateFile('templates/quizimage_details.pt')
    fields = field.Fields(interfaces.IQuizImage).select('title', 'contents')
    mode = DISPLAY_MODE

    @property
    def can_modify(self):
        person = IPerson(self.request.principal, None)
        return person in self.context.editors

    def edit_url(self):
        context_url = absoluteURL(self.context, self.request)
        return '%s/edit.html' % context_url


class QuizImageViewDoneLinkViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
        <h3 i18n:domain="schooltool" class="done-link">
          <a tal:attributes="href view/url"
             i18n:translate="">
            Done
          </a>
        </h3>
    ''')

    def url(self):
        person = IPerson(self.request.principal, None)
        if person is not None:
            return '%s/quiz_images.html' % (absoluteURL(person, self.request))


class PersonQuizImagesView(flourish.page.Page, PersonSections):

    implements(interfaces.ISchoolToolQuizView,
               interfaces.IPersonQuizView)

    container_class = 'container widecontainer'

    @Lazy
    def person(self):
        return self.context

    @Lazy
    def quizImageContainer(self):
        app = ISchoolToolApplication(None)
        return interfaces.IQuizImageContainer(app)

    @Lazy
    def editable_quiz_images(self):
        result = []
        for image in self.quizImageContainer.values():
            if self.person in image.editors:
                result.append(image)
        return result

    def update(self):
        container = self.quizImageContainer
        if not flourish.canEdit(container):
            raise NotFound(self.context, self.__name__, self.request)


class PersonQuizImageAddLink(QuizImageAddLink, PersonSections):

    @Lazy
    def person(self):
        return IPerson(self.request.principal, None)

    @property
    def enabled(self):
        if self.person is not None:
            return self.instructor_sections
        return False

    @property
    def url(self):
        app = ISchoolToolApplication(None)
        container = interfaces.IQuizImageContainer(app)
        camefrom = '%s/quiz_images.html' % absoluteURL(self.context,
                                                       self.request)
        return '%s/add.html?camefrom=%s' % (
            absoluteURL(container, self.request), camefrom)


class EditableQuizImagesViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <tal:block i18n:domain="schooltool.quiz"
                 tal:define="images view/view/editable_quiz_images">
        <h3 i18n:translate="">Editable Images</h3>
        <div tal:condition="images"
             tal:define="ajax nocall:view/view/providers/ajax"
             tal:content="structure ajax/view/context/editable_quiz_images_table"
             />
        <p i18n:translate="" tal:condition="not:images">
          There are no images you can edit yet.
        </p>
      </tal:block>
    ''')



def image_preview_formatter(value, item, formatter):
    params = {
        'title': item.title,
        'url': '%s/inline' % absoluteURL(item, formatter.request),
        }
    return '<img width="300px" src="%(url)s" alt="%(title)s"/>' % params


class EditableQuizImagesTable(table.ajax.Table):

    def items(self):
        return self.view.editable_quiz_images

    def columns(self):
        default = super(EditableQuizImagesTable, self).columns()
        created = GetterColumn(
            name='created',
            title=_('Created'),
            getter=lambda i, f: IZopeDublinCore(i).created,
            cell_formatter=lambda v, i, f: render_date(v),
            subsort=True)
        preview = GetterColumn(
            name='preview',
            title=_('Preview'),
            cell_formatter=image_preview_formatter)
        directlyProvides(created, ISortableColumn)
        return default + [created, preview]

    def sortOn(self):
        return (('created', True),)

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data editable-quiz-images-table'})


class QuizImageInsertLink(flourish.page.ModalFormLinkViewlet):

    @property
    def dialog_title(self):
        return _('Insert Image')

    @property
    def url(self):
        url = super(QuizImageInsertLink, self).url
        return '%s?item_id=%s' % (url, self.view.item.__name__)



class QuizItemAddViewQuizImageInsertLink(flourish.page.ModalFormLinkViewlet):

    @property
    def dialog_title(self):
        return _('Insert Image')


class InsertQuizImageDialog(flourish.form.Dialog):

    template = InlineViewPageTemplate('''
      <div tal:content="structure view/context/schooltool:content/ajax/view/container/insert_images_table" />
    ''')

    @Lazy
    def container(self):
        app = ISchoolToolApplication(None)
        return interfaces.IQuizImageContainer(app)

    def update(self):
        flourish.form.Dialog.update(self)

    def initDialog(self):
        super(InsertQuizImageDialog, self).initDialog()
        self.ajax_settings['dialog']['height'] = '500'
        self.ajax_settings['dialog']['width'] = '900'


def title_cell_formatter(value, item, formatter):
    item_url = absoluteURL(item, formatter.request)
    url = '%s/quiz_images/%s/inline' % (APP_ROOT_PATTERN, item.__name__)
    item_id = formatter.request.get('item_id', '')
    link_item_url = '%s/link_item.html?item_id=%s' % (item_url, item_id)
    return '<a onclick="return ST.quiz.insert_image(this);" href="%s" title="%s" class="%s" id="%s">%s</a>' % (url, value, link_item_url, item.__name__, item.title)


class InsertQuizImagesTable(table.ajax.Table):

    def columns(self):
        default = super(InsertQuizImagesTable, self).columns()
        created = GetterColumn(
            name='created',
            title=_('Created'),
            getter=lambda i, f: IZopeDublinCore(i).created,
            cell_formatter=lambda v, i, f: render_date(v),
            subsort=True)
        preview = GetterColumn(
            name='preview',
            title=_('Preview'),
            cell_formatter=image_preview_formatter)
        directlyProvides(created, ISortableColumn)
        return default + [created, preview]

    def sortOn(self):
        return (('created', True),)

    def updateFormatter(self):
        if self._table_formatter is not None:
            return
        self.setUp(formatters=[title_cell_formatter],
                   table_formatter=self.table_formatter,
                   batch_size=self.batch_size,
                   prefix=self.__name__,
                   css_classes={'table': 'data quiz-images-table editable-quiz-images-table'})


class InsertQuizImagesTableFilter(table.ajax.TableFilter):

    template = ViewPageTemplateFile('templates/quiz_images_table_filter.pt')


class InsertQuizImagesTableHiddenFields(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <input type="hidden"
             tal:define="value view/field_value"
             tal:condition="value"
             tal:attributes="name view/field_name;
                             value value;"
             />
    ''')

    @property
    def field_name(self):
        return 'item_id'

    @property
    def field_value(self):
        return self.request.get(self.field_name)


class DownloadExistingImagesView(flourish.page.Page):

    @Lazy
    def item_container(self):
        return interfaces.IQuizItemContainer(self.context)

    @Lazy
    def image_container(self):
        return interfaces.IQuizImageContainer(self.context)

    def __call__(self):
        chooser = INameChooser(self.image_container)
        downloaded = {}
        for item in self.item_container.values():
            quizzes = list(item.quizzes)
            if quizzes:
                if self.hasImage(item):
                    url = self.getImageURL(item)
                    if self.isExternalURL(url):
                        editors = set()
                        for quiz in quizzes:
                            for editor in quiz.editors:
                                editors.add(removeSecurityProxy(editor))
                        if url not in downloaded:
                            title = self.getImageTitle(item) or url
                            image = self.createImage(url, title)
                            name = chooser.chooseName('', image)
                            self.image_container[name] = image
                            notify(ObjectCreatedEvent(image))
                            downloaded[url] = name
                        else:
                            name = downloaded[url]
                            container = self.image_container
                            image = removeSecurityProxy(container[name])
                        image_url = '%s/inline' % absoluteURL(image,
                                                              self.request)
                        self.replaceExternalURL(item, url, image_url)
                        image.items.add(removeSecurityProxy(item))
                        for editor in editors:
                            if editor not in image.editors:
                                image.editors.add(editor)
        self.request.response.redirect(self.nextURL())

    def hasImage(self, item):
        return '.. image::' in item.body

    def getImageURL(self, item):
        for line in item.body.splitlines():
            if '.. image::' in line:
                return line.split('::')[-1].strip()

    def getImageTitle(self, item):
        for line in item.body.splitlines():
            if ':alt:' in line:
                return line.split(':alt:')[-1].strip()

    def isExternalURL(self, url):
        app_url = self.nextURL()
        return not url.startswith(app_url)

    def createImage(self, url, title):
        result = QuizImage()
        result.title = title
        data = urlopen(url).read()
        try:
            image = Image.open(StringIO(data))
            if image.format not in ('JPEG', 'PNG'):
                raise IOError()
        except (IOError,):
            raise ValueError('%s is not a JPEG or PNG image' % url)
        data = self.processImage(image)
        f = ImageFile()
        self.updateFile(f, data)
        result.contents = f
        return result

    def processImage(self, image):
        kw = {'quality': 100, 'filter': Image.ANTIALIAS}
        result = image
        f = StringIO()
        self._format = image.format
        result.save(f, self._format, **kw)
        return f.getvalue()

    def updateFile(self, ob, data):
        ob.mimeType = Image.MIME[self._format]
        w = ob.open("w")
        w.write(data)
        w.close()

    def replaceExternalURL(self, item, url, new_url):
        item.body = item.body.replace(url, new_url)

    def nextURL(self):
        app = ISchoolToolApplication(None)
        return absoluteURL(app, self.request)


class LinkItemView(flourish.page.Page):

    @Lazy
    def item_container(self):
        app = ISchoolToolApplication(None)
        return interfaces.IQuizItemContainer(app)

    def __call__(self):
        response = self.request.response
        response.setHeader('Content-Type', 'application/json')
        encoder = flourish.tal.JSONEncoder()
        result = False
        item_id = self.request.get('item_id')
        if item_id is not None:
            item = self.item_container.get(item_id)
            if item is not None:
                item = removeSecurityProxy(item)
                removeSecurityProxy(self.context).items.add(item)
                result = True
        json = encoder.encode(result)
        return json


class QuizImageItemsView(EditRelationships):

    implements(interfaces.ISchoolToolQuizView,
               interfaces.IMathJaxView)

    container_class = 'container widecontainer'

    current_title = _('Current items')
    available_title = _('Add items')

    def getAvailableItemsContainer(self):
        app = ISchoolToolApplication(None)
        return interfaces.IQuizItemContainer(app)

    @property
    def title(self):
        return self.context.title

    def getCollection(self):
        return self.context.items


class ItemsTable(table.ajax.Table):

    def columns(self):
        body = ReStructuredTextColumn(
            name='body',
            title=_('Body'),
            schema_field=interfaces.IQuizItem['body']
            )
        type_ = TypeColumn(
            name='type',
            title=_('Type'))
        return [body, type_]

    def sortOn(self):
        return None


class ItemAddRelationshipTable(RelationshipAddTableMixin,
                               ItemsTable):
    pass


class ItemRemoveRelationshipTable(RelationshipRemoveTableMixin,
                                  ItemsTable):
    pass


class QuizImageDeleteLinkViewlet(flourish.page.ModalFormLinkViewlet):

    @property
    def dialog_title(self):
        title = _('Delete this quiz image?')
        return translate(title, context=self.request)

    @property
    def enabled(self):
        return not bool(list(self.context.items))


class QuizImageContainerDeleteView(flourish.containers.ContainerDeleteView):

    def nextURL(self):
        if 'CONFIRM' in self.request:
            person = IPerson(self.request.principal)
            url = '%s/quiz_images.html' % absoluteURL(person, self.request)
            return url
        return super(QuizImageContainerDeleteView, self).nextURL()


class QuizImageDeleteView(flourish.form.DialogForm):

    template = ViewPageTemplateFile('templates/confirm_delete_quiz_image.pt')

    dialog_submit_actions = ('delete',)
    dialog_close_actions = ('cancel',)
    label = None

    @button.buttonAndHandler(_('Delete'), name='delete')
    def handle_delete(self, action):
        url = '%s/delete.html?delete.%s&CONFIRM' % (
            absoluteURL(self.context.__parent__, self.request),
            self.context.__name__)
        self.request.response.redirect(url)
        self.ajax_settings['dialog'] = 'close'

    @button.buttonAndHandler(_('Cancel'), name='cancel')
    def handle_cancel(self, action):
        pass

    def updateActions(self):
        super(QuizImageDeleteView, self).updateActions()
        self.actions['delete'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')
