#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation,
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""
SchoolTool Quiz quiz item views.
"""

from urllib import urlencode
from zope import schema
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.component import adapts, getUtility
from zope.container.interfaces import INameChooser
from zope.event import notify
from zope.i18n.interfaces.locales import ICollator
from zope.interface import implements, Interface, directlyProvides
from zope.lifecycleevent import ObjectCreatedEvent
from zope.publisher.interfaces import NotFound
from zope.schema import ValidationError
from zope.schema import TextLine
from zope.schema.interfaces import IVocabularyFactory
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.absoluteurl import absoluteURL

from z3c.form import validator, widget
from z3c.form import field, form, button
from z3c.form.browser.radio import RadioFieldWidget
from z3c.form.interfaces import IObjectWidget, DISPLAY_MODE, HIDDEN_MODE
from z3c.form.object import ObjectSubForm
from z3c.form.object import SubformAdapter
from z3c.form.object import registerFactoryAdapter
import zc.resourcelibrary

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.cando.interfaces import ICourseSkill
from schooltool.cando.interfaces import ICourseSkills
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.course.interfaces import IInstructor
from schooltool.person.interfaces import IPerson
from schooltool.skin import flourish
from schooltool.term.interfaces import ITerm, IDateManager

from schooltool.quiz import QuizMessage as _
from schooltool.quiz import interfaces
from schooltool.quiz.browser.widget import LabeledCheckBoxFieldWidget
from schooltool.quiz.quizitem import Choice
from schooltool.quiz.quizitem import QuizItem

registerFactoryAdapter(interfaces.IChoice, Choice)


class OneCorrectChoiceRequired(ValidationError):

    __doc__ = _('One correct choice is required.')


class SolutionRequiredError(ValidationError):

    __doc__ = _('Required input is missing.')


DoNotShowChoiceLabel = widget.StaticWidgetAttribute(
    False,
    field=interfaces.IQuizItem['choices']
    )


class ChoicesValidator(validator.SimpleFieldValidator):

    def validate(self, value):
        super(ChoicesValidator, self).validate(value)
        if not self.view.is_open_question():
            if not value:
                raise OneCorrectChoiceRequired(value)
            correct = []
            for choice in value:
                if choice.correct:
                    correct.append(choice)
            if not correct:
                raise OneCorrectChoiceRequired(value)


class SolutionValidator(validator.SimpleFieldValidator):

    def validate(self, value):
        super(SolutionValidator, self).validate(value)
        if self.view.is_open_question() and not value:
            raise SolutionRequiredError(value)


class ChoiceSubForm(ObjectSubForm):

    def setupFields(self):
        super(ChoiceSubForm, self).setupFields()
        self.fields['correct'].widgetFactory = LabeledCheckBoxFieldWidget


def getCourse(skill, item):
    if ICourseSkill.providedBy(skill):
        return skill.__parent__.__parent__.__parent__
    elif item is not None:
        return item.course


class SkillVocabulary(SimpleVocabulary):

    def __init__(self, terms, item=None, *interfaces):
        self.by_value = {}
        self.by_token = {}
        self._terms = terms
        self.item = item
        for term in self._terms:
            value = (getCourse(term.value, self.item), term.value)
            if value in self.by_value:
                raise ValueError(
                    'term values must be unique: %s' % repr(value))
            if term.token in self.by_token:
                raise ValueError(
                    'term tokens must be unique: %s' % repr(term.token))
            self.by_value[value] = term
            self.by_token[term.token] = term
        if interfaces:
            directlyProvides(self, *interfaces)

    def getSkillWithCourse(self, value):
        course = getCourse(value, self.item)
        courseskills = ICourseSkills(course)
        courseskillset = courseskills.get(value.__parent__.__name__, {})
        courseskill = courseskillset.get(value.__name__)
        return (course, courseskill)

    def __contains__(self, value):
        """See zope.schema.interfaces.IBaseVocabulary"""
        try:
            return self.getSkillWithCourse(value) in self.by_value
        except TypeError:
            # sometimes values are not hashable
            return False

    def getTerm(self, value):
        """See zope.schema.interfaces.IBaseVocabulary"""
        try:
            return self.by_value[self.getSkillWithCourse(value)]
        except KeyError:
            raise LookupError(value)


class QuizItemForm(flourish.form.Form):

    implements(interfaces.ISchoolToolQuizView)

    id = 'quiz-item-form'
    legend = _('Question Information')
    container_class = 'container widecontainer'
    content_template = ViewPageTemplateFile('templates/form.pt')
    _open_question_token = 'open'

    @property
    def fields(self):
        result = field.Fields(interfaces.IQuizItem)
        result = result.select('type_', 'body', 'choices', 'solution')
        result += field.Fields(self.course_field)
        result += field.Fields(self.skill_field)
        result['type_'].widgetFactory = RadioFieldWidget
        return result

    @Lazy
    def item_container(self):
        app = ISchoolToolApplication(None)
        return interfaces.IQuizItemContainer(app)

    def encodeWidget(self, widget):
        name = widget.name
        value = widget.value
        if isinstance(value, list):
            # XXX: this assumes single item lists being passed
            if len(value) == 1:
                value = value[0]
        return {name: value}

    def encodeWidgets(self):
        result = {}
        for widget in self.widgets.values():
            result.update(self.encodeWidget(widget))
        if self.is_open_question():
            solution_widget = self.widgets['solution']
            result.update(self.encodeWidget(solution_widget))
        else:
            choices_widget = self.widgets['choices']
            for widget in choices_widget.widgets:
                subform = widget.subform
                for subwidget in subform.widgets.values():
                    result.update(self.encodeWidget(subwidget))
            choices_count = '%s.count' % choices_widget.name
            result[choices_count] = len(choices_widget.widgets)
        return result

    def is_open_question(self):
        item_type = getattr(self.widgets['type_'], 'value', [])
        return self._open_question_token in item_type

    def setQuiz(self, item):
        quiz = removeSecurityProxy(self.context)
        quiz.order.append(item.__name__)
        item.quizzes.add(quiz)

    @button.buttonAndHandler(_('Submit'), name='submit')
    def handle_submit(self, action):
        self._process_submit(action)

    @button.buttonAndHandler(_('Submit and add'), name='submitadd')
    def handle_submit_add(self, action):
        self._process_submit_add(action)

    @button.buttonAndHandler(_('Submit and copy'), name='submitcopy')
    def handle_submit_copy(self, action):
        self._process_submit_copy(action)

    @button.buttonAndHandler(_('Cancel'), name='cancel')
    def handle_cancel(self, action):
        self._process_cancel(action)

    def get_errors(self):
        return self.widgets.errors

    def updateWidgets(self):
        super(QuizItemForm, self).updateWidgets()
        widget = self.widgets['choices']
        widget.addClass('choices-widget')
        for name, action in widget.actions.items():
            action.addClass('button-cancel')
            if name in ('add'):
                action.title = _('Add choice')
            elif name in ('remove'):
                action.title = _('Remove selected choices')

    @property
    def courses(self):
        result = set()
        current_term = getUtility(IDateManager).current_term
        person = IPerson(self.request.principal, None)
        if person is not None:
            for section in list(IInstructor(person).sections()):
                term = ITerm(section)
                if term == current_term:
                    for course in section.courses:
                        if ICourseSkills(course):
                            result.add(course)
        return result

    @property
    def course_vocabulary(self):
        collator = ICollator(self.request.locale)
        terms = []
        for course in sorted(self.courses,
                              key=lambda course: course.title,
                              cmp=collator.cmp):
            term = SimpleTerm(removeSecurityProxy(course),
                              token=course.__name__,
                              title=course.title)
            terms.append(term)
        return SimpleVocabulary(terms)

    @property
    def course_field(self):
        return schema.Choice(
            __name__='course',
            title=_('Skill Course'),
            vocabulary=self.course_vocabulary,
            required=False)

    @property
    def skill_vocabulary(self):
        collator = ICollator(self.request.locale)
        terms = []
        for course in self.courses:
            worksheets = ICourseSkills(course)
            for worksheet in worksheets.values():
                for skill in sorted(worksheet.values(),
                                    key=lambda skill: skill.label,
                                    cmp=collator.cmp):
                    token = self.getSkillId(skill)
                    title = self.getSkillTitle(skill)
                    term = SimpleTerm(removeSecurityProxy(skill),
                                      token=token,
                                      title=title)
                    terms.append(term)
        return SkillVocabulary(terms, getattr(self, 'item', None))

    @property
    def skill_field(self):
        return schema.Choice(
            __name__='skill',
            title=_('Skill'),
            vocabulary=self.skill_vocabulary,
            required=False)

    def getSkillId(self, skill):
        worksheet = skill.__parent__
        worksheets = worksheet.__parent__
        course = worksheets.__parent__
        return '%s.%s.%s' % (course.__name__,
                             worksheet.__name__,
                             skill.__name__)

    def getSkillTitle(self, skill):
        worksheet = skill.__parent__
        return '%s / %s: %s' % (worksheet.title,
                                skill.label,
                                skill.title)


class QuizItemAddView(QuizItemForm, form.AddForm):

    _created_obj = None
    _nextURL = None
    _encodeWidgets = None
    _finishedAdd = False

    @property
    def fields(self):
        result = super(QuizItemAddView, self).fields
        result += field.Fields(
            TextLine(__name__='image', title=u'Image ID', required=False),
            mode=HIDDEN_MODE)
        return result

    def create(self, data):
        item = QuizItem()
        self._image = data.pop('image')
        form.applyChanges(self, item, data)
        return item

    def add(self, item):
        item_container = self.item_container
        chooser = INameChooser(item_container)
        name = chooser.chooseName('', item)
        item_container[name] = item
        self.setQuiz(item)
        self.setImage(item)
        self._item = item
        self._finishedAdd = True
        return item

    def setImage(self, item):
        if self._image is not None:
            app = ISchoolToolApplication(None)
            images = interfaces.IQuizImageContainer(app)
            image = images.get(self._image)
            if image is not None:
                image = removeSecurityProxy(image)
                if item not in image.items:
                    image.items.add(item)
        
    def update(self):
        super(QuizItemAddView, self).update()
        errors = self.get_errors()
        if errors:
            self.errors = errors
            self.status = self.formErrorsMessage
        elif self._created_obj is not None:
            self.add(self._created_obj)
            if self._encodeWidgets:
                params = self.encodeWidgets()
                self._nextURL = '%s?%s' % (self.action, urlencode(params))

    def nextURL(self):
        return self._nextURL

    def updateActions(self):
        super(QuizItemAddView, self).updateActions()
        self.actions['submit'].addClass('button-ok')
        self.actions['submitadd'].addClass('button-cancel')
        self.actions['submitcopy'].addClass('button-cancel')
        self.actions['cancel'].addClass('button-cancel')

    def get_created_obj(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        return self.create(data)

    def _process_submit(self, action):
        self._created_obj = self.get_created_obj(action)
        self._nextURL = absoluteURL(self.context, self.request)

    def _process_submit_add(self, action):
        self._created_obj = self.get_created_obj(action)
        self._nextURL = self.action

    def _process_submit_copy(self, action):
        self._created_obj = self.get_created_obj(action)
        self._encodeWidgets = True

    def _process_cancel(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def updateWidgets(self):
        super(QuizItemAddView, self).updateWidgets()
        self.widgets['course'].prompt = True
        self.widgets['course'].promptMessage = _('Select a course')
        self.widgets['skill'].prompt = True
        self.widgets['skill'].promptMessage = _('Select a skill')


class ChoiceSubformAdapter(SubformAdapter):

    adapts(Interface,
           flourish.IFlourishLayer,
           Interface,
           QuizItemForm,
           IObjectWidget,
           Interface,
           Interface)

    factory = ChoiceSubForm


validator.WidgetValidatorDiscriminators(
    ChoicesValidator,
    view=QuizItemForm,
    field=interfaces.IQuizItem['choices'])


validator.WidgetValidatorDiscriminators(
    SolutionValidator,
    view=QuizItemForm,
    field=interfaces.IQuizItem['solution'])


class QuizItemViewBase(flourish.page.Page):

    @Lazy
    def item_container(self):
        app = ISchoolToolApplication(None)
        return interfaces.IQuizItemContainer(app)

    @Lazy
    def item(self):
        item_id = self.request.get('item_id')
        item = self.item_container.get(item_id)
        if item in self.context.items:
            return item
        raise NotFound(self.context, item_id, self.request)

    @Lazy
    def type_token(self):
        factory = getUtility(IVocabularyFactory,
                             name='schooltool.quiz.quiz_item_types')
        vocabulary = factory(None)
        return vocabulary.getTerm(self.item.type_).token

    @property
    def can_modify(self):
        person = IPerson(self.request.principal, None)
        return person in self.context.editors


class QuizItemView(QuizItemViewBase):

    implements(interfaces.ISchoolToolQuizView,
               interfaces.IMathJaxView)

    container_class = 'container widecontainer'

    @property
    def title(self):
        return self.context.title

    @property
    def subtitle(self):
        result = {
            'selection': _('Selection Question'),
            'multiple': _('Multiple Selection Question'),
            'open': _('Open Question'),
            }
        return result[self.type_token]


class QuizItemEditView(QuizItemForm, form.EditForm, QuizItemViewBase):

    @property
    def title(self):
        result = {
            'selection': _('Selection Question'),
            'multiple': _('Multiple Selection Question'),
            'open': _('Open Question'),
            }
        return result[self.type_token]

    def update(self):
        zc.resourcelibrary.need('schooltool.table')
        return form.EditForm.update(self)

    def updateActions(self):
        super(QuizItemEditView, self).updateActions()
        self.actions['submit'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def nextURL(self):
        url = self.request.get('camefrom',
                               absoluteURL(self.context, self.request))
        return url

    @button.buttonAndHandler(_('Submit'), name='submit')
    def handle_submit(self, action):
        super(QuizItemEditView, self).handleApply.func(self, action)
        if (self.status == self.successMessage or
            self.status == self.noChangesMessage):
            self.request.response.redirect(self.nextURL())

    @button.buttonAndHandler(_('Cancel'), name='cancel')
    def handle_cancel(self, action):
        self.request.response.redirect(self.nextURL())

    def getContent(self):
        return self.item

    def updateWidgets(self):
        super(QuizItemEditView, self).updateWidgets()
        self.widgets['course'].prompt = True
        self.widgets['course'].promptMessage = _('No course')
        self.widgets['skill'].prompt = True
        self.widgets['skill'].promptMessage = _('No skill')


class QuizItemDuplicateView(QuizItemViewBase):

    def __call__(self):
        copy = self.item.copy()
        chooser = INameChooser(self.item_container)
        name = chooser.chooseName('', copy)
        self.item_container[name] = copy
        copy.quizzes.add(removeSecurityProxy(self.context))
        removeSecurityProxy(self.context).order.append(name)
        notify(ObjectCreatedEvent(copy))
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)


class QuizItemRemoveView(flourish.form.DialogForm, QuizItemViewBase):

    template = ViewPageTemplateFile('templates/confirm_remove_item.pt')

    dialog_submit_actions = ('remove',)
    dialog_close_actions = ('cancel',)
    label = None

    @button.buttonAndHandler(_('Remove'), name='remove')
    def handle_remove(self, action):
        removeSecurityProxy(self.context.items).remove(self.item)
        removeSecurityProxy(self.context).order.remove(self.item.__name__)
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)
        self.ajax_settings['dialog'] = 'close'

    @button.buttonAndHandler(_('Cancel'), name='cancel')
    def handle_cancel(self, action):
        pass

    def updateActions(self):
        super(QuizItemRemoveView, self).updateActions()
        self.actions['remove'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class QuizItemFormScriptViewlet(flourish.viewlet.Viewlet):

    template = ViewPageTemplateFile('templates/quiz_item_form_script.pt')


class QuizItemViewTitleViewlet(flourish.viewlet.Viewlet):

    template = ViewPageTemplateFile('templates/quiz_item_title.pt')

    def edit_url(self):
        quiz_url = absoluteURL(self.context, self.request)
        camefrom = '%s/view_item.html?item_id=%s' % (
            quiz_url, self.view.item.__name__)
        return '%s/edit_item.html?item_id=%s&camefrom=%s' % (
            quiz_url, self.view.item.__name__, camefrom)


class QuizItemViewBodyViewlet(flourish.viewlet.Viewlet):

    template = ViewPageTemplateFile('templates/quiz_item_body.pt')

    def render_rst(self, value, iface, field_name):
        renderer = flourish.form.Form(value, self.request)
        renderer.mode = DISPLAY_MODE
        renderer.fields = field.Fields(iface[field_name])
        renderer.update()
        widget = renderer.widgets[field_name]
        return widget.render()

    def body(self):
        value = self.view.item
        return self.render_rst(value, interfaces.IQuizItem, 'body')

    def solution(self):
        value = self.view.item
        return self.render_rst(value, interfaces.IQuizItem, 'solution')

    def choices(self):
        for choice in self.view.item.choices:
            yield {
                'body': self.render_rst(choice, interfaces.IChoice, 'body'),
                'correct': choice.correct,
                }

    def is_open_question(self):
        return self.view.type_token == 'open'

    def has_skill(self):
        return None not in (self.view.item.course, self.view.item.skill)


class QuizItemViewDoneLinkViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
        <h3 i18n:domain="schooltool" class="done-link">
          <a tal:attributes="href context/@@absolute_url"
             i18n:translate="">
            Done
          </a>
        </h3>
    ''')
