#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation,
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""
SchoolTool Quiz widget views.
"""

from cStringIO import StringIO
from zope.cachedescriptors.property import Lazy
from zope.component import adapts, getUtility
from zope.i18n import translate
from zope.i18n.interfaces.locales import ICollator
from zope.interface import implements, Interface
from zope.schema.interfaces import IDatetime
from zope.schema.interfaces import IVocabularyFactory
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.absoluteurl import absoluteURL

from z3c.form import field
from z3c.form.browser import widget
from z3c.form.browser.checkbox import CheckBoxWidget
from z3c.form.browser.checkbox import SingleCheckBoxWidget
from z3c.form.browser.radio import RadioWidget
from z3c.form.browser.text import TextWidget
from z3c.form.browser.textarea import TextAreaWidget
from z3c.form.converter import BaseDataConverter
from z3c.form.converter import FormatterValidationError
from z3c.form.interfaces import DISPLAY_MODE
from z3c.form.widget import FieldWidget, Widget
from zc.table.column import GetterColumn

try:
    import Image
except ImportError:
    from PIL import Image

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.common import parse_datetime
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.course.interfaces import IInstructor, ILearner
from schooltool.skin import flourish
from schooltool import table

from schooltool.quiz import QuizMessage as _
from schooltool.quiz import interfaces
from schooltool.quiz.utils import render_rst, render_pre


class IDatetimeTextWidget(Interface):
    """Custom date time widget"""


class IReStructuredTextWidget(Interface):
    """Marker for ReStructuredText widgets"""


class ReStructuredTextWidget(TextAreaWidget):

    implements(IReStructuredTextWidget)

    klass = u'textarea-widget restructuredtext-widget'

    def update(self):
        Widget.update(self)
        widget.addFieldClass(self)
        if self.required and self.mode != DISPLAY_MODE:
            self.addClass('required')

    def render_rst(self):
        return render_rst(self.value, self.request)


class PreLabeledCheckBoxWidget(CheckBoxWidget):

    def render_pre(self, label):
        return render_pre(label)


class RestLabeledCheckBoxWidget(CheckBoxWidget):

    def render_rst(self, label):
        return render_rst(label, self.request)


class RestLabeledRadioWidget(RadioWidget):

    def render_rst(self, label):
        return render_rst(label, self.request)


def ReStructuredTextFieldWidget(field, request):
    return FieldWidget(field, ReStructuredTextWidget(request))


def LabeledCheckBoxFieldWidget(field, request):
    # the default z3c.form implementation hides the widget's label
    return FieldWidget(field, SingleCheckBoxWidget(request))


def PreLabeledCheckBoxFieldWidget(field, request):
    return FieldWidget(field, PreLabeledCheckBoxWidget(request))


def RestLabeledCheckBoxFieldWidget(field, request):
    return FieldWidget(field, RestLabeledCheckBoxWidget(request))


def RestLabeledRadioFieldWidget(field, request):
    return FieldWidget(field, RestLabeledRadioWidget(request))


class SimpleFormAdapter(object):

    def __init__(self, context):
        self.__dict__['context'] = removeSecurityProxy(context)

    def __setattr__(self, name, value):
        setattr(self.context, name, value)

    def __getattr__(self, name):
        return getattr(self.context, name)


class CustomDatetimeTextWidget(TextWidget):

    implements(IDatetimeTextWidget)


def CustomDatetimeFieldTextWidget(field, request):
    return FieldWidget(field, CustomDatetimeTextWidget(request))


class CustomDatetimeDataConverter(BaseDataConverter):

    adapts(IDatetime, CustomDatetimeTextWidget)

    def toWidgetValue(self, value):
        if value is self.field.missing_value:
            return u''
        return value.strftime('%Y-%m-%d %H:%M:00')

    def toFieldValue(self, value):
        if value == u'':
            return self.field.missing_value
        try:
            value = parse_datetime(value)
        except (ValueError,):
            raise FormatterValidationError(
                _('The datetime string did not match the pattern '
                  'YYYY-MM-DD HH:MM:SS'),
                value)
        try:
            value.strftime('%Y-%m-%d')
        except (ValueError,):
            raise FormatterValidationError(
                _('Year has to be equal or greater than 1900'),
                value)
        return value


# XXX: Core should detect the format if None

class ImageDataConverter(flourish.widgets.ImageDataConverter):

    adapts(interfaces.IImage, flourish.widgets.IImageWidget)

    def processImage(self, image):
        kw = {'quality': 100, 'filter': Image.ANTIALIAS}
        result = image
        if (self.field.size is not None and
            image.size != self.field.size):
            size = self.getMaxSize(image)
            image.thumbnail(size, kw['filter'])
            if image.size != self.field.size or image.mode == 'RGBA':
                result = Image.new('RGB', self.field.size, (255, 255, 255))
                left = int(round((self.field.size[0] - image.size[0])/float(2)))
                up = int(round((self.field.size[1] - image.size[1])/float(2)))
                mask = None
                if image.mode == 'RGBA':
                    mask = image
                result.paste(image, (left, up), mask)
        f = StringIO()
        if self.field.format is None:
            self._format = image.format
        else:
            self._format = self.field.format
        result.save(f, self._format, **kw)
        return f.getvalue()

    def updateFile(self, ob, data):
        ob.mimeType = Image.MIME[self._format]
        w = ob.open("w")
        w.write(data)
        w.close()


class PersonSections(object):

    person = None

    @Lazy
    def learner_sections(self):
        return list(ILearner(self.person).sections())

    @Lazy
    def instructor_sections(self):
        return list(IInstructor(self.person).sections())


class PersonQuizViewTertiaryNavigationManager(
    flourish.page.TertiaryNavigationManager):

    template = InlineViewPageTemplate("""
        <ul tal:attributes="class view/list_class">
          <li tal:repeat="item view/items"
              tal:attributes="class item/class"
              tal:content="structure item/viewlet">
          </li>
        </ul>
    """)

    @Lazy
    def quizImageContainer(self):
        app = ISchoolToolApplication(None)
        return interfaces.IQuizImageContainer(app)

    @property
    def items(self):
        result = []
        path = self.request['PATH_INFO']
        current = path[path.rfind('/')+1:]
        actions = [
            ('quizzes.html', _('Quizzes')),
            ]
        if flourish.canEdit(self.quizImageContainer):
            actions.append(
            ('quiz_images.html', _('Images'))
            )
        for action, title in actions:
            url = '%s/%s' % (absoluteURL(self.context, self.request), action)
            title = translate(title, context=self.request)
            result.append({
                'class': action == current and 'active' or None,
                'viewlet': u'<a href="%s">%s</a>' % (url, title),
                })
        return result


def editors_formatter(editors, item, formatter):
    collator = ICollator(formatter.request.locale)
    result = sorted([person.title for person in editors],
                    cmp=collator.cmp)
    return '<br />'.join(result)


class ReStructuredTextColumn(GetterColumn):

    def __init__(self, *args, **kw):
        self.schema_field = kw.pop('schema_field')
        super(ReStructuredTextColumn, self).__init__(*args, **kw)

    def renderCell(self, item, formatter):
        renderer = flourish.form.Form(item, formatter.request)
        renderer.mode = DISPLAY_MODE
        renderer.fields = field.Fields(self.schema_field)
        renderer.update()
        return renderer.widgets['body'].render()


class TypeColumn(GetterColumn):

    def __init__(self, *args, **kw):
        super(TypeColumn, self).__init__(*args, **kw)
        factory = getUtility(IVocabularyFactory,
                             name='schooltool.quiz.quiz_item_types')
        vocabulary = factory(None)
        self.vocabulary = vocabulary

    def getter(self, item, formatter):
        term = self.vocabulary.getTerm(item.type_)
        return term.title
