#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation,
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""
SchoolTool Quiz interfaces.
"""

from zope import schema
from zope.container.constraints import contains
from zope.container.interfaces import IContainer, IContained
from zope.interface import Attribute, Interface, implements

from schooltool.requirement.interfaces import IRequirement
from schooltool.skin.flourish import fields

from schooltool.quiz.utils import DATETIME_FORMAT
from schooltool.quiz import QuizMessage as _


class IQuizContainer(IContainer):

    contains('.IQuiz')


class IQuizItemContainer(IContainer):

    contains('.IQuizItem')


class IQuiz(IContained):

    order = schema.List(title=u'Order of items')

    title = schema.TextLine(title=_('Title'))

    items = Attribute('IQuizItem objects (see IRelationshipProperty)')

    editors = Attribute('IBasicPerson objects (see IRelationshipProperty)')

    deployed = Attribute('IDeployedQuiz objects (see IRelationshipProperty)')

    def copy():
        """Returns a new Quiz"""

    def updateOrder(item_id, position):
        """Updates order of the items"""

    def __iter__():
        """Iterates through items"""


class IReStructuredText(Interface):
    """Marker for fields that render ReStructuredText"""


class ReStructuredText(schema.Text):

    implements(IReStructuredText)


class IChoice(Interface):

    body = ReStructuredText(title=_('Body'))

    correct = schema.Bool(title=_('Correct?'), default=False)


class IQuizItem(IRequirement):

    body = ReStructuredText(title=_('Body'))

    type_ = schema.Choice(
        title=_('Type'),
        vocabulary='schooltool.quiz.quiz_item_types')

    solution = ReStructuredText(title=_('Solution'), required=False)

    choices = schema.List(
        title=_('Choices'),
        value_type=schema.Object(title=_('Choice'), schema=IChoice),
        required=False)

    quizzes = Attribute('IQuiz objects (see IRelationshipProperty)')

    deployed = Attribute('IQuiz objects (see IRelationshipProperty)')

    course = Attribute('ICourse object')

    skill = Attribute('ISkill object')

    images = Attribute('IQuizImage objects (see IRelationshipProperty)')

    def copy():
        """Returns a new QuizItem"""


class IDeployedQuizContainer(IContainer):

    contains('.IDeployedQuiz')


class IDeployedQuiz(IContained):

    order = schema.List(title=u'Order of items')

    title = schema.TextLine(title=_('Title'))

    random = schema.Bool(title=_('Random'), default=False)

    start = schema.Datetime(title=_('Start'), description=DATETIME_FORMAT)

    end = schema.Datetime(title=_('End'), description=DATETIME_FORMAT)

    open_by_teacher = schema.Bool(title=_('Open'), default=False)

    quizzes = Attribute('IQuiz object (see IRelationshipProperty)')

    items = Attribute('IQuizItem objects (see IRelationshipProperty)')

    worksheet_id = schema.TextLine(title=_('Worksheet ID'), required=False)

    activity_id = schema.TextLine(title=_('Activity ID'), required=False)

    period_id = schema.TextLine(title=_('Period ID'), required=False)

    def __iter__():
        """Iterates through items"""


class IDeployedQuizData(Interface):
    """Metadata for IDeployedQuiz objects"""


class IDeployedQuizGradebookData(IDeployedQuizData):

    updated = schema.Datetime(title=_('Grades last update'), required=False)

    worksheet_id = schema.TextLine(title=_('Worksheet ID'), required=False)

    activity_id = schema.TextLine(title=_('Activity ID'), required=False)


class ISchoolToolQuizView(Interface):
    """Marker for views that need the SchoolTool Quiz library"""


class IMathJaxView(Interface):
    """Marker for views that need MathJax CDN"""

class IRule(Interface):
    """Schooltool Quiz Cando Interface Settings"""
 
    level = schema.Int(title=_("Skill Level"))

    total_taken = schema.Int(title=_("Total Taken"))

    number_correct = schema.Int(title=_("Number Correct"))

    out_of = schema.Int(title=_("Out Of Most Recent"))



class IApplicationPreferences(Interface):
    """SchoolTool Quiz preferences annotation adapter"""

    mathjax_help = ReStructuredText(
        title=_('Quiz Editing Help'),
        required=False)

    skill_rules = schema.List(title = _("Skill Level Rules"),
                              description=_('The "Skill Level" is automatically set as a score for a skill if one of two conditions are met: 1) the student answered a "Total Taken" number of items or 2) the student answered correctly a "Number Correct" of items out of the last "Out of Most Recent". Setting 0 as the value of any of the three criteria fields means UNLIMITED'),
                              value_type = schema.Object(schema=IRule))

    save_time = schema.Float(title= _("Autosave Interval (In Milliseconds)"),
                             required=True)
    
class IQuizImageContainer(IContainer):

    contains('.IQuizImage')

class IImage(fields.IImage):

    pass


class Image(fields.Image):

    implements(IImage)

    format = None


class IQuizImage(IContained):

    title = schema.TextLine(title=_('Title'))

    contents = Image(title=_('Contents'))

    items = Attribute('IQuizItem objects (see IRelationshipProperty)')

    editors = Attribute('IBasicPerson objects (see IRelationshipProperty)')


class IPersonQuizView(Interface):
    """Marker for person views that needs tertiary navigation"""

class IHaveAssignments(Interface):
    """Marker for objects which can have assignments"""

class IAssignments(IContainer):
    contains(".IAssignment")
    """Container for assigned quiz items"""

    def addAssignment(self, assignment):
        """adds assignments to the container"""

    def get(self, key):
        """gets values based on key"""
 

class IAssignment(IContained):
    """Assigned items""" 
