#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation,
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""
SchoolTool Quiz quizitem.
"""

from persistent import Persistent
from zope.cachedescriptors.property import Lazy
from zope.component import adapter, adapts
from zope.container.btree import BTreeContainer
from zope.interface import implements, implementer
from zope.lifecycleevent.interfaces import IObjectRemovedEvent
from zope.proxy import sameProxiedObjects
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm

from schooltool.app.app import InitBase, StartUpBase
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.cando.interfaces import ICourseSkillSet
from schooltool.relationship import RelationshipProperty
from schooltool.requirement.requirement import Requirement
from schooltool.schoolyear.subscriber import ObjectEventAdapterSubscriber

from schooltool.quiz import QuizMessage as _
from schooltool.quiz import interfaces
from schooltool.quiz import relationships


APP_QUIZITEMCONTAINER_KEY = 'schooltool.quiz.quiz_item'


class QuizItemContainer(BTreeContainer):

    implements(interfaces.IQuizItemContainer)


class QuizItem(Requirement):

    implements(interfaces.IQuizItem)

    body = None
    type_ = None
    solution = None
    choices = None
    course = None
    skill = None

    quizzes = RelationshipProperty(relationships.URIQuizItems,
                                   relationships.URIQuizItem,
                                   relationships.URIQuiz)

    images = RelationshipProperty(relationships.URIQuizItemImages,
                                  relationships.URIQuizItem,
                                  relationships.URIQuizImage)

    def __init__(self):
        Requirement.__init__(self, None)

    def copy(self):
        result = QuizItem()
        result.body = self.body
        result.type_ = self.type_
        result.solution = self.solution
        result.course = self.course
        result.skill = self.skill
        if self.choices is not None:
            result_choices = []
            for choice in self.choices:
                result_choice = Choice()
                result_choice.body = choice.body
                result_choice.correct = choice.correct
                result_choices.append(result_choice)
            result.choices = result_choices
        for image in self.images:
            result.images.add(image)
        return result


class Choice(Persistent):

    implements(interfaces.IChoice)

    body = None
    correct = None

    def __eq__(self, other):
        return other is not None and \
               self.body == other.body and \
               self.correct == other.correct


def QuizItemTypeVocabulary(context):
    items = [
        ('selection', _('Selection')),
        ('multiple', _('Multiple Selection')),
        ('open', _('Open')),
        ]
    terms = [SimpleTerm(item, token=token, title=item)
             for token, item in items]
    return SimpleVocabulary(terms)


def QuizItemTypeVocabularyFactory():
    return QuizItemTypeVocabulary


class QuizItemContainerInit(InitBase):

    def __call__(self):
        self.app[APP_QUIZITEMCONTAINER_KEY] = QuizItemContainer()


class QuizItemContainerAppStartup(StartUpBase):

    def __call__(self):
        if APP_QUIZITEMCONTAINER_KEY not in self.app:
            self.app[APP_QUIZITEMCONTAINER_KEY] = QuizItemContainer()


@implementer(interfaces.IQuizItemContainer)
@adapter(ISchoolToolApplication)
def getApplicationQuizItemContainer(app):
    return app[APP_QUIZITEMCONTAINER_KEY]


class CourseWorksheetRemoved(ObjectEventAdapterSubscriber):

    adapts(IObjectRemovedEvent, ICourseSkillSet)

    def __call__(self):
        skillset = self.object
        for skill in skillset.values():
            for item in self.quiz_items:
                if sameProxiedObjects(item.skill, skill):
                    item.course = None
                    item.skill = None

    @Lazy
    def quiz_items(self):
        app = ISchoolToolApplication(None)
        quiz_container = interfaces.IQuizItemContainer(app)
        return quiz_container.values()
